/****************************************************************************************/
/* Ardunino Watchdog                                                                    */
/* Version: 1.00                                                                        */
/* Date: 24.04.2020                                                                     */
/* Compiled with Arduino 1.8.5                                                          */
/* Arduino board: ATtiny 13                                                             */
/* Arduino processor: ATtiny 13A                                                        */
/* Clock 1Mhz (Intern)                                                                  */
/* Programmer --> Arduio als ISP                                                        */
/*                                                                                      */
/* Description:https://arduino-projekte.webnode.at/attiny-programmierung/               */
/*        https://www.kollino.de/arduino/attiny-support-unter-arduino-1-6-installieren/ */
/*                                                                                      */
/* Pin 1 = PB5 = Reseved for reset                                                      */
/* Pin 2 = PB3 = Watchdog Trigger input                                                 */
/* Pin 3 = PB4 = Watchdog disable input                                                 */
/* Pin 4 = GND                                                                          */
/* Pin 5 = PB0 = NC                                                                     */
/* Pin 6 = PB1 = Power output active high --> Input pin FET                             */
/* Pin 7 = PB2 = Reset ESP32 --> Via 4k7 resistor and NPN OC transsistor to EN pin      */
/*                                                                                      */
/****************************************************************************************/


/****************************************************************************************/
/* Definitions                                                                          */
/****************************************************************************************/
const int WD_DISABLE   = 4;              /* Pin 1 = PB5 = Watchdog disable, high = normal operation, low = disable */
const int WDT_INPUT    = 3;              /* Pin 2 = PB3 = Watchdog trigger input */
const int RESET_OUTPUT = 2;              /* Pin 7 = PB2 = Reset output active high */
const int POWER_OUTPUT = 1;              /* Pin 6 = PB1 = Switch power on and off active high */
const int WD_TIMER = 8000;               /* 8000ms = 8 sec */
const int POWER_ON_DELAY = 10000;        /* Power on delay after power on */
const int RESET_PULSE_LENGTH = 2000;     /* Reset pulse length = 100ms */
const int MAX_INIT_DELAY = 600;          /* 600 seconds = 10 Min max init delay */
const byte INIT_DELAY = 20;              /* Init delay before WD waits for trigger = 20 Seconds */

/****************************************************************************************/
/* Variables                                                                            */
/****************************************************************************************/
boolean PULSE_DETECT;                    /* False = Waiting for positive pulse, True = Waiting for negative pulse */
byte LOOP_COUNTER;                       /* Loop counter */
unsigned int WD_COUNTER;                 /* WD counter */
unsigned int INIT_COUNTER;               /* Init counter */


/****************************************************************************************/
/* Setup procedure                                                                      */
/****************************************************************************************/ 
void setup() 
{
  pinMode (POWER_OUTPUT, OUTPUT);        /* Configure power output as output */
  digitalWrite(POWER_OUTPUT, LOW);       /* Set power output to Low --> Switcvh power off */
  pinMode (RESET_OUTPUT, OUTPUT);        /* Configure power output as output */
  digitalWrite(RESET_OUTPUT, LOW);       /* Set reset output to Low --> Power off */ 
  pinMode (WDT_INPUT, INPUT);            /* Configure WD trigger as input */
  pinMode (WD_DISABLE, INPUT);           /* Configure WD disable as input */

  if (digitalRead (WD_DISABLE) == 1)     /* WD not disabled ? */
  {
    SUB_RESET_SEQUENCE();                /* Call sub routine for reset sequence */
  }
  else                                   /* WD disabled ? */
  {
    digitalWrite(POWER_OUTPUT, HIGH);    /* Set reset output to High --> Power on */
    delay (100);                         /* 100ms delay */
    digitalWrite(RESET_OUTPUT, HIGH);    /* Set reset output to High --> Reset */
    delay (RESET_PULSE_LENGTH);          /* Reset delay */
    digitalWrite(RESET_OUTPUT, LOW);     /* Set reset output to Low --> No reset */ 
    WD_COUNTER = 0;                      /* Reset WD counter */ 
    PULSE_DETECT = false;                /* Waiting for positive pulse */
  }
}

/****************************************************************************************/
/* Main loop                                                                            */
/****************************************************************************************/
void loop() 
{
  delay (1);                             /* One ms delay */
  if (digitalRead (WD_DISABLE) == 1)     /* WD not disabled ? */
  {
    WD_COUNTER = WD_COUNTER + 1;         /* Increase watchdog counter */
  }
  else                                   /* WD disabled ? */
  {
    WD_COUNTER = 0;                      /* Reset WD_COUNTER */
  }
      
  /* Positive edge detected after WD input was low ? */
  if ((digitalRead (WDT_INPUT) == 1) && (PULSE_DETECT == false))      
  {
    WD_COUNTER = 0;                      /* Reset WD_COUNTER */
    PULSE_DETECT = true;                 /* Set pulse detect true = Wait for signal going low */
  }

  /* Negative edge detected after WD input was high ? */
  if ((digitalRead (WDT_INPUT) == 0) && (PULSE_DETECT == true))  
  {
    WD_COUNTER = 0;                      /* Reset WD_COUNTER */    
    PULSE_DETECT = false;                /* Set pulse detect true = Wait for next positive edge */
  }     

  if (WD_COUNTER >= WD_TIMER)            /* Watchdog has expired ? */
  {
    SUB_RESET_SEQUENCE();                 /* Call sub routine for reset sequence */
  }
}


/****************************************************************************************/
/* Subroutine for reset sequence                                                        */
/****************************************************************************************/
void SUB_RESET_SEQUENCE (void)
{
  digitalWrite (POWER_OUTPUT, LOW);      /* Set power output to Low --> Switcvh power off */
  delay (POWER_ON_DELAY);                /* Power on off delay */
  digitalWrite(POWER_OUTPUT, HIGH);      /* Set reset output to High --> Power on */
  delay (100);                           /* 100ms delay */
  digitalWrite(RESET_OUTPUT, HIGH);      /* Set reset output to High --> Reset */
  delay (RESET_PULSE_LENGTH);            /* Reset delay */
  digitalWrite(RESET_OUTPUT, LOW);       /* Set reset output to Low --> No reset */  

  /* Intitial delay before WD gets active */
  for (LOOP_COUNTER=0; LOOP_COUNTER < INIT_DELAY; LOOP_COUNTER++)
  {
    delay (1000);                        /* 1 sec delay */
  }

  /* Wait for first negative pulse to trigger watchdog or max init delay reached */
  INIT_COUNTER = 0;                      /* Reset init counter */
  LOOP_COUNTER = 0;                      /* Reset loop counter */
  do
  {
    delay (10);                          /* 10 ms delay */
    LOOP_COUNTER++;                      /* Increase loop counter */
    if (LOOP_COUNTER >= 100)             /* 1000ms = 1 sec reached */
    {
      LOOP_COUNTER = 0;                  /* Reset loop counter */
      INIT_COUNTER ++;                   /* Increase init counter */
    }
  }
  while ((digitalRead (WDT_INPUT) == 1) && (INIT_COUNTER <= MAX_INIT_DELAY));
  PULSE_DETECT = false;                  /* Waiting for positive pulse */
  WD_COUNTER = 0;                        /* Reset WD counter */
}
