/****************************************************************************************/
/* LAN Neo Pixel remote lightening for garden with NTP based on ESP32                   */
/* Version: 2.4                                                                         */
/* Board: OLIMEX ESP32-GATEWAY --> Achtung auf Revision, da sonst Absturz!!!            */
/* Board revision: E                                                                    */
/* Date: 20.10.2020                                                                     */
/* Compiled with Arduino 1.8.12                                                         */
/* ESP32 arduino extension: 1.0.4                                                       */
/*                                                                                      */
/* Arduino OTA: https://lastminuteengineers.com/esp32-ota-updates-arduino-ide/          */
/* NTP integration:                                                                     */
/*         https://lastminuteengineers.com/esp32-ntp-server-date-time-tutorial/         */
/*         https://github.com/espressif/arduino-esp32/blob/master/cores/                */
/*           esp32/esp32-hal-time.c#L47                                                 */
/*                                                                                      */
/* FLASH handling: https://randomnerdtutorials.com/esp32-flash-memory/                  */
/* ESP Async Web Server:                                                                */
/*   https://techtutorialsx.com/2017/12/01/esp32-arduino-asynchronous-http-webserver/   */
/*                                                                                      */
/* Libraries:                                                                           */
/*            ESP32 Ping: https://github.com/marian-craciunescu/ESP32Ping               */
/*            Time library: https://github.com/PaulStoffregen/Time                      */
/*            ESP32 Async Web Server: https://github.com/me-no-dev/ESPAsyncWebServer    */
/*            Async TCP: https://github.com/me-no-dev/AsyncTCP                          */
/*            ESP32 DMX Library: https://github.com/claudeheintz/LXESP32DMX             */
/*                                                                                      */
/* Other tools:                                                                         */
/*        ESP32 Filesystem Uploader:                                                    */
/*        https://randomnerdtutorials.com/install-esp32-filesystem-uploader-arduino-ide */
/*                                                                                      */
/*                                                                                      */
/* Usage of serial interfaces:                                                          */
/* Serial for debug interface                                                           */
/* Pin asignments:                                                                      */
/*    GPIO32 = DMX Control Pin                                                          */
/*    GPIO16 = DMX Transmit Pin                                                         */
/*    GPIO33  = Watchdog trigger                                                        */
/*                                                                                      */
/* INIT.TXT: Init file with basic access parameters which must be uploaded via          */
/*           ESP32 Sketch Data Upload                                                   */
/*      File content:                                                                   */
/*           Line 1:  Hostname                                                          */
/*           Line 2:  Fixed IP address, e.g. 192.168.123.221                            */
/*           Line 3:  Subnet mask, e.g. 255.255.255.0                                   */
/*           Line 4:  Gateway IP address, e.g. 192.168.123.1                            */
/*           Line 5:  DNS Server 1, e.g. 195.58.161.123                                 */
/*           Line 6:  DNS Server 2, e.g. 212.186.211.21                                 */
/*           Line 7:  NTP server 1, e.g. time.nist.gov                                  */
/*           Line 8:  NTP server 2, e.g. pool.ntp.org                                   */
/*           Line 9:  NTP server 3, e.g. europe.pool.ntp.org                            */
/*      Warning: No ENTER (CR/LF) allowed after line 9 !!!                              */  
/*                                                                                      */
/* UDP                                                                                  */
/* IP: 192.168.123.221, Port: 6454                                                      */
/* Message Format 41 72 74 2d 4e 65 74 00 00 50 00 0e 00 00 00 00 00 03 00 05 ff        */
/* 41 72 74 2d 4e 65 74 = Art-Net                                                       */
/* 00 00 = Trennzeichen                                                                 */ 
/* 50 00 = Art DMX datapacket                                                           */
/* 0e 00 = Artnet Version Number                                                        */
/* 00 = Squence                                                                         */
/* 00 = Physical                                                                        */
/* 00 00 = Universe                                                                     */
/* 03 = Message length                                                                  */
/* 00 = DMX channel 1 value                                                             */
/* 05 = DMX channel 2 value                                                             */
/* ff = DMX channel 3 value                                                             */
/*                                                                                      */
/* HTTP                                                                                 */
/* IP: 192.168.123.221, Port: 80                                                        */
/* Message Format Send                                                                  */
/*  http://192.168.123.221/?login=<LOGIN>:<PASSWORD>/c=<D0><D1><D2>                     */
/*  <D0>-<D2> = Command 0 - 255                                                         */
/*  Example for power on: http://192.168.123.221/?login=admin:pt230567/c=129            */
/*                                                                                      */
/* 000x xxxx               Color, Stores and Programs                                   */
/*                         1-13 Single color, 15-19 Storage, 20-29 Programs             */
/*                                                                                      */
/*                         1 = Single color cold white                                  */
/*                         2 = Single color warm white                                  */
/*                         3 = Single color yellow                                      */
/*                         4 = Single color orange                                      */
/*                         5 = Single color red                                         */
/*                         6 = Single color dark purple                                 */
/*                         7 = Single color purple                                      */
/*                         8 = Single color bright purple                               */
/*                         9 = Single color blue                                        */
/*                        10 = Single color bright blue                                 */
/*                        11 = Single color cyan                                        */
/*                        12 = Single color bright green                                */
/*                        13 = Single color green                                       */
/*                        15 = Storage position 1                                       */
/*                        16 = Storage position 2                                       */
/*                        17 = Storage position 3                                       */
/*                        18 = Storage position 4                                       */
/*                        19 = Storage position 5                                       */
/*                        20 = Program 0: Change 1: 3 Second hold, 3 Second change      */
/*                                                  White - Red - White - Green -       */
/*                                                  White - Blue - White - Yellow -     */
/*                                                  White - Cyan - White - Purple       */
/*                        21 = Program 1: Change 2: 6 Seconds hold, 0 Seconds change    */
/*                                                  Red - Green - Blue - Yellow -       */
/*                                                  Purple - Yellow - Cyan - Purple     */
/*                        22 = Program 2: Change 3: 3 Seconds hold, 3 Seconds change    */
/*                                                  Red - Green - Blue - Yellow         */
/*                                                  Puple - Yellow - Cyan - Purple      */
/*                        23 = Program 3: Change 4: 0 Seconds hold, 6 Second change     */
/*                                                  Red - Orange - Yellow - Light green */
/*                                                  Green - Cyan - Blue - Light Purple  */
/*                                                  Purple - Dark purple                */
/*                        24 = Program 4: Change 5: Spektrum 0 Sec. hold, 6 Sec. change */
/*                                                  Yellow - Red -  Purple - Blue       */
/*                                                  Cyan - Green                        */
/*                        25 = Program 5: Steps: Red Up/down, Green Up/Down,            */
/*                                               Blue Up/Down                           */
/*                        26 = Program 6: Color Flash 1: Green - Red - Blue             */
/*                                                     Yellow - Cyan - Purple           */
/*                                                     Green - Blue - Purple            */
/*                                                     Green - Orange - White           */
/*                                                     Cyan - Yellow - White            */
/*                        27 = Program 7: Color Flash 2: Yellow - Red - Cyan - Yellow   */
/*                                                       Cyan - Purple - Green - Blue   */
/*                                                       Purple - Green - Orange - White*/
/*                                                       Green - Yellow - Purple        */
/*                        28 = Program 8: Fireplace: Yellow Orange Up/Down              */
/*                        29 = Program 9: Thunderstorm: Blue flashing                   */
/*                                                                                      */
/* 0010 xxxx               32-40: Dimmer values 0-8 for single colors                   */
/* 0011 xxxx               48-51: Dimmer values 0-3 for programs                        */
/* 0100 xxxx               64-73: Speed values 0-9 for programs                         */
/* 1000 xxxx               Special commands                                             */
/*                                 0000 Off (128)                                       */
/*                                 0001 On (129)                                        */
/*                                 0010 Pause (130)                                     */
/*                                 0011 Play (131)                                      */
/*                                 0101 Report status (133)                             */
/*                                      Feedback: a;b;c;d:e:f;g                         */
/*                                      a: Power On/Off Status: On/Off                  */
/*                                      b: Actual time/date: XX:XX, XX.XX.20XX          */
/*                                      c: Actual On Time: XX:XX                        */
/*                                      d: Modus flag 0/1: Off/Auto                     */
/*                                      e: Slave 1 modus flag 0/1: Off/Auto             */
/*                                      f: Slave 2 modus flag 0/1: Off/Auto             */
/*                                      g: Slave 3 modus flag 0/1: Off/Auto             */
/*                                 1000 Store value in storage position 1 (136)         */
/*                                 1001 Store value in storage position 2 (137)         */
/*                                 1010 Store value in storage position 3 (138)         */
/*                                 1011 Store value in storage position 4 (139)         */
/*                                 1100 Store value in storage position 5 (140)         */
/*                                 1101 Mode = 0 = Off (141)                            */
/*                                 1110 Mode = 1 = Auto (142)                           */                               
/*                                 1111 Restore default values in storage 1-5 (143)     */
/*                                                                                      */
/* 1001 xxxx               Special commands slave on/off                                */
/*                                 0000 Slave 1 = Mode = 0 = Off (144)                  */
/*                                 0001 Slave 1 = Mode = 1 = Auto (145)                 */ 
/*                                 0010 Slave 2 = Mode = 0 = Off (146)                  */
/*                                 0011 Slave 2 = Mode = 1 = Auto (147)                 */ 
/*                                 0100 Slave 3 = Mode = 0 = Off (148)                  */
/*                                 0101 Slave 3 = Mode = 1 = Auto (149)                 */  
/*                         Special command bootloader                                   */
/*                                 0110 Enter bootloader (150)                          */ 
/*                         Special command reset                                        */
/*                                 0111 Reset device (151)                              */  
/* FLASH structure                                                                      */
/*                        0 = Actual program (1-13, 15-19, 20-29)                       */
/*                        1 = Color dimming factor (0-8) 0=no dimming, 8=max dimming    */
/*                        2 = Program dimming factor (0-3)                              */
/*                            0 = no dimming --> Stored value 0                         */
/*                            1 = dim step 1 --> Stored value 63                        */
/*                            2 = dim step 2 --> Stored value 126                       */
/*                            3 = dim step 3 --> Stored value 189                       */
/*                        3 = Program speed (0-9), 0=highest, 9=lowest                  */
/*                        4 = Automatic Mode: 0 = Off, 1 = Auto                         */
/*                        5 = Slave 1 automatic Mode: 0 = Off, 1 = Auto                 */
/*                        7 = Slave 2 automatic Mode: 0 = Off, 1 = Auto                 */
/*                        9 = Slave 3 automatic Mode: 0 = Off, 1 = Auto                 */
/*                        10-13 Storage position 1                                      */
/*                              10 = Red color value                                    */
/*                              11 = Green color value                                  */
/*                              12 = Blue color value                                   */
/*                              13 = Color dimming (0-9) 0=no dimming, 8=max dimming    */
/*                        20-23 Storage position 2                                      */
/*                              20 = Red color value                                    */
/*                              21 = Green color value                                  */
/*                              22 = Blue color value                                   */
/*                              23 = Color dimming (0-9) 0=no dimming, 8=max dimming    */
/*                        30-33 Storage position 3                                      */
/*                              30 = Red color value                                    */
/*                              31 = Green color value                                  */
/*                              32 = Blue color value                                   */
/*                              33 = Color dimming (0-9) 0=no dimming, 8=max dimming    */
/*                        40-43 Storage position 4                                      */
/*                              40 = Red color value                                    */
/*                              41 = Green color value                                  */
/*                              42 = Blue color value                                   */
/*                              43 = Color dimming (0-9) 0=no dimming, 8=max dimming    */
/*                        50-53 Storage position 5                                      */
/*                              50 = Red color value                                    */
/*                              51 = Green color value                                  */
/*                              52 = Blue color value                                   */
/*                              53 = Color dimming (0-9) 0=no dimming, 8=max dimming    */
/*                        60 = REMOTE UPDATE VIA LAN                                    */
/*                              0 = Normal boot                                         */
/*                              1 = Remote update                                       */
/*                        61 = NTP UPDATE HOUR                                          */
/*                              0-23                                                    */
/*                        62 = NTP UPDATE MIN                                           */
/*                              0-59                                                    */
/*                        63 = NTP TIME ZONE                                            */
/*                              0 = - 5 hours                                           */
/*                              1 = - 4 hours                                           */
/*                              2 = - 3 hours                                           */
/*                              3 = - 2 hours                                           */
/*                              4 = - 1 hour                                            */
/*                              5 =   0 hours                                           */
/*                              6 = + 1 hour                                            */
/*                              7 = + 2 hours                                           */
/*                              8 = + 3 hours                                           */
/*                              9 = + 4 hours                                           */
/*                             10 = + 5 hours                                           */
/*                                                                                      */
/*                        64 = NTP autiotic summer/winter switching                     */
/*                              0 = No automatic summertime switching                   */
/*                              1 = Automatic summertime switching                      */
/*                                                                                      */
/* Display duing boot up:                                                               */
/*         Initialize SPIFFS:                                                           */
/*                    Failed: Fast blinking red 0,5 sec                                 */
/*                    Done: One yellow blink for 2 seconds                              */
/*         Connect to LAN:                                                              */
/*                    Failed: One red blink for 2 seconds before watchdog reset         */
/*                    Done: One purple blink for 2 seconds                              */
/*         Connect to NTP:                                                              */
/*                    Failed: One red blink for 2 seconds before watchdog reset         */
/*                    Done: One cyan blink for 2 seconds                                */
/*         Started server and configuration:                                            */
/*                    Failed: One red blink for 2 seconds before watchdog reset         */
/*                    Done: One green blink for 2 seconds                               */
/****************************************************************************************/

/****************************************************************************************/
/* Libraries                                                                            */
/****************************************************************************************/
#include <EEPROM.h>                               /* Include EEPROM read/write */
#include <ETH.h>                                  /* Include Ethernet library */
#include <WiFiUDP.h>                              /* Include ESP32 WIFI UDPlibrary */
#include "LXESP32DMX.h"                           /* Include ESP32 DMX library */
#include <ESP32Ping.h>                            /* Include ESP32 ping library */
#include "SPIFFS.h"                               /* Include ESP32 library to access files */
#include <Time.h>                                 /* Uses time library */
#include <ESPAsyncWebServer.h>                    /* Include ESP32 Async Web Server librarby */ 
/* Libraries for OTA */
#include <ESPmDNS.h>                              /* Include ESP DNS library */
#include <ArduinoOTA.h>                           /* Include OTA library */


/****************************************************************************************/
/* Definitions of hardware pins                                                         */
/****************************************************************************************/
const byte DMX_DIRECTION_PIN     = 32;            /* Define DMX pin for direction, High = Send, Low = Receive */
const byte DMX_SERIAL_OUTPUT_PIN = 16;            /* Define DMX pin for sending data */
const byte WD_TRIGGER            = 33;            /* Define pin for watchdog trigger */


/****************************************************************************************/
/* Definitions for general settings                                                     */
/****************************************************************************************/
/* If this is uncommented here, status messages will be output via the serial interface, if commented out then disabled ! */
//#define _DEBUG_                                   
#define EEPROM_SIZE 70                            /* Define the number of bytes of FLASH to be used */
uint8_t dmxbuffer[DMX_MAX_FRAME];                 /* Define DMX buffer */


/****************************************************************************************/
/* Definitions for NTP server                                                           */
/****************************************************************************************/
boolean NTP_CONNECT;                              /* False = NTP not received OK, True = NTP received OK */
boolean NTP_SUMMERTIME;                           /* False = No automatic summertime switching, True = Automatic summertime switching */
byte    NTP_UPDATE_HOUR;                          /* Hour of time to update the time via NTP */
byte    NTP_UPDATE_MIN;                           /* Minute of time to update the time via NTP */
byte    NTP_TIME_ZONE;                            /* NTP time zone: 0 = - 5 hours, 5 = 0 hour, 10 = +5 hours */
int     DAYLIGHT_OFFSET_SEC = 0;                  /* Preset daylight offset to wintertime in seconds */
long    GMT_OFFSET_SEC;                           /* Calculated from NTP_TIME_ZONE, e.g. Austia = UTC + 1 hour = 3600 seconds */
char    NTP_SERVER_1 [21] = {0};                  /* String for NTP Server 1 */
char    NTP_SERVER_2 [21] = {0};                  /* String for NTP Server 2 */
char    NTP_SERVER_3 [21] = {0};                  /* String for NTP Server 3 */
/* Define time zone values */
const int NTP_TIME_ZONE_VALUE[] = {-5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5};


/****************************************************************************************/
/* Definitions for time library and time handling                                       */
/****************************************************************************************/
byte ACT_SEC;                                     /* Actual Second */
byte NEW_SEC;                                     /* New second */
byte ACT_MIN;                                     /* Actual minute */
byte ACT_HOUR;                                    /* Actual hour */
byte ACT_DAY;                                     /* Actual day */
byte ACT_MONTH;                                   /* Actual month */
byte ACT_YEAR_SHORT;                              /* Actual year 2 digit */
int  ACT_YEAR;                                    /* Actual year 4 digit */
struct tm TIMEINFO;                               /* Time structure */


/****************************************************************************************/
/* Definitions for Ethernet settings and handling                                       */
/****************************************************************************************/
boolean ETH_CONNECT;                              /* True = LAN connected, False = LAN not connected */
char HOSTNAME [21] = {0};                         /* String for hostname */
IPAddress FIXED_IP;                               /* IP address for fixed IP */
IPAddress SUBNET;                                 /* Subnet mask */
IPAddress GATEWAY;                                /* IP address for gateway address */
IPAddress DNS_1;                                  /* IP adress for DNS Server 1 */
IPAddress DNS_2;                                  /* IP adress for DNS Server 2 */


/****************************************************************************************/
/* Definitions for Webserver                                                            */
/****************************************************************************************/
AsyncWebServer server(80);                        /* Create AsyncWebServer object on port 80 */
WiFiUDP udpclient;                                /* Define UDP client as udpclient */


/****************************************************************************************/
/* Definitions for web page and IP request handling                                     */
/****************************************************************************************/
const char* KEYWORD = "login";                    /* Define keyword to be checked */
const char* LOGIN = "admin:password/";            /* Define login and password to be checked */ 


/****************************************************************************************/
/* Constants                                                                            */
/****************************************************************************************/
/* Sun down time for Vienna/Austria for month January - December minus 1*/
const byte SUN_DOWN_HOUR[] = {15,16,17,18,19,19,19,19,18,16,15,15};
const byte SUN_DOWN_MIN[]  = {31,16,32,47,27,53,35,4,5,35,19,17};
const byte RECEIVE_BUFFER_SIZE = 160;             /* 160 byte receiver buffer */
const byte TEMP_BUFFER_SIZE = 160;                /* 160 byte temporarly buffer */
const int UDP_PORT = 6454;                        /* Define UDP Port */
const byte NO_DIM_STEPS=200;                      /* Number of dim steps per sequence */
const char* SW_VERSION = "2.4"; 

/* Sequenz array for program pointer */
/* This table is the pointer to the different programs */
/* for the table PROG_DATA, e.g. program 1 points to   */
/* 0, program 2 points to 61, ...                      */
const int PROG_POINTER[] = {0,61,102,143,194,225,346,497,648,749};

/* Sequenz array for dimming steps in percent */
/* e.g. 100 = 100%, 60 = 60%                  */
const int DIM_CONST[] = {100,60,30,15};

/* Sequenz array for dimming steps in pervcent */
/*e.g. 100 = 3 sec, 200 = 6 sec, 10 = 0,3 sec  */
const int SPEED_CONST[] = {6, 10, 25, 50, 75, 100, 200, 400, 800, 1600};

/* Sequenz array for program */
/* The dataset contains the following information */
/* The first value defines the number of steps    */
/* Then each data set consists of 5 values        */
/*     First value: Red value                     */
/*     Second value: Green value                  */
/*     Third value: Blue value                    */
/*     Forth value: Seconds for change            */
/*     Fifth value: Seconds for hold              */
const unsigned char PROG_DATA[] PROGMEM = {
12,
255,255,126,3,3,
255,0,0,3,3,
255,255,126,3,3,
0,255,0,3,3,
255,255,126,3,3,
0,0,255,3,3,
255,255,126,3,3,
255,255,0,3,3,
255,255,126,3,3,
0,255,255,3,3,
255,255,126,3,3,
255,0,255,3,3,
8,            
255,0,0,6,0,
0,255,0,6,0,
0,0,255,6,0,
255,255,0,6,0,
255,0,255,6,0,
255,255,0,6,0,
0,255,255,6,0,
255,0,255,6,0,
8,          
255,0,0,3,3,
0,255,0,3,3,
0,0,255,3,3,
252,252,0,3,3,
255,0,255,3,3,
255,255,0,3,3,
0,255,255,3,3,
255,0,255,3,3,
10,
255,0,0,6,0,
255,72,0,6,0,
255,255,0,6,0,
92,255,0,6,0,
0,255,0,6,0,
0,54,255,6,0,
0,0,255,6,0,
90,0,255,6,0,
255,0,255,6,0,
255,0,90,6,0,
6,      
255,255,0,6,0,
255,0,0,6,0,
255,0,255,6,0,
0,0,255,6,0,
0,255,255,6,0,
0,255,0,6,0,
24, 
0,0,0,4,0,   
18,0,0,4,2,
9,0,0,4,2,
36,0,0,4,2,
9,0,0,4,2,
92,0,0,4,2,
9,0,0,4,2,
255,0,0,4,2,
0,0,0,4,0,
0,18,0,4,2,
0,9,0,4,2,
0,36,0,4,2,
0,9,0,4,2,
0,92,0,4,2,
0,9,0,4,2,
0,255,0,4,2,
0,0,0,4,0,
0,0,18,4,2,
0,0,9,4,2,
0,0,36,4,2,
0,0,9,4,2,
0,0,92,4,2,
0,0,9,4,2,
0,0,255,4,3,
30,
0,252,0,0,3,
0,0,0,0,3,
252,0,0,0,3,
0,0,0,0,3,
0,0,252,0,3,
0,0,0,0,3,
252,252,0,0,3,
0,0,0,0,3,
0,252,252,0,3,
0,0,0,0,3,
252,0,252,0,3,
0,0,0,0,3,
0,252,0,0,3,
0,0,0,0,3,
0,0,252,0,3,
0,0,0,0,3,
252,0,252,0,3,
0,0,0,0,3,
0,252,0,0,3,
0,0,0,0,3,
252,35,0,0,3,
0,0,0,0,2,
252,252,252,0,3,
0,0,0,0,3,
0,252,252,0,3,
0,0,0,0,3,
252,252,0,0,3,
0,0,0,0,3,
252,252,252,0,3,
0,0,0,0,3,
30,
144,252,0,0,4,
0,0,0,0,2,
252,0,0,0,4,
0,0,0,0,2,
0,117,252,0,4,
0,0,0,0,2,
252,99,0,0,4,
0,0,0,0,2,
0,252,117,0,4,
0,0,0,0,2,
126,0,252,0,4,
0,0,0,0,2,
0,252,0,0,4,
0,0,0,0,2,
0,0,252,0,4,
0,0,0,0,2,
252,0,63,0,4,
0,0,0,0,2,
0,252,108,0,4,
0,0,0,0,2,
252,18,0,0,4,
0,0,0,0,2,
252,252,252,0,4,
0,0,0,0,2,
0,252,144,0,4,
0,0,0,0,2,
252,144,16,0,4,
0,0,0,0,2,
36,0,252,0,4,
0,0,0,0,2,
20,
252,36,0,2,3,
252,18,0,0,2,
252,36,0,0,2,
252,45,0,3,3,
252,36,0,0,3,
252,18,0,2,0,
252,45,0,4,2,
252,18,0,0,2,
252,45,0,6,4,
252,54,0,0,2,
252,45,0,2,3,
252,36,0,0,3,
252,18,0,0,2,
252,36,0,3,2,
252,18,0,0,2,
252,45,0,0,3,
252,18,0,3,0,
252,36,0,0,2,
252,45,0,3,0,
252,18,0,0,2,
18,
0,0,0,0,3,
0,27,171,0,1,
0,0,0,0,2,
0,117,252,0,2,
0,252,252,4,1,
0,252,117,2,1,
0,54,252,0,3,
0,0,0,0,3,
0,0,63,0,1,
0,0,126,0,1,
0,0,189,0,1,
0,54,252,0,1,
0,0,0,0,2,
0,63,126,2,1,
0,117,252,2,1,
0,252,252,2,1,
0,252,117,2,1,
0,126,63,2,1
};

/****************************************************************************************/
/* Variables                                                                            */
/****************************************************************************************/
boolean POWER_OFF;                                /* True when standby activated */
boolean POWER_WEB_OFF;                            /* True when standby activated to show correct information on web page */
boolean UDP_COLOR = false;                        /* UDP color select, True when selected */
boolean RESET_FLAG;                               /* False = No reset, True = Reset */
boolean WD_FLAG;                                  /* True = High, False = Low */
boolean BOOTLOADER_FLAG;                          /* False = Do not enter bootloader, True = Enter bootloader */
boolean SLAVE_1_CHANGED;                          /* False = No change, true = Slave 1 mode changed via web interface */
boolean SLAVE_2_CHANGED;                          /* False = No change, true = Slave 2 mode changed via web interface */
boolean SLAVE_3_CHANGED;                          /* False = No change, true = Slave 3 mode changed via web interface */
boolean TEST_COLOR;                               /* False = No test color, True = Test color */
boolean SUB_MENU_FLAG;                            /* Flag for sub menu 4; False = Display storage settings with 0, true = Load actual storage settings after recall */
byte ON_TIME_HOUR;                                /* On time hour */
byte ON_TIME_MIN;                                 /* On time minute */
byte TEST_POS_1;                                  /* Test position 1 for general calculation */
byte ACT_PROGRAM;                                 /* Actual program */
byte COLOR_DIM_STEP;                              /* Actual color dimming factor */
byte PROG_DIM_STEP;                               /* Actual program dimming factor */
byte PROG_SPEED;                                  /* Program speed */
byte MAX_RED_COLOR;                               /* Maximum dimmer value red color */
byte MAX_GREEN_COLOR;                             /* Maximum dimmer value green color */
byte MAX_BLUE_COLOR;                              /* Maximum dimmer value blue color */
byte RED_DIM;                                     /* Red color dimmer value */
byte GREEN_DIM;                                   /* Green color dimmer value */
byte RED_COLOR;                                   /* Red color value */
byte GREEN_COLOR;                                 /* Green color value */
byte BLUE_COLOR;                                  /* Blue color value */
byte BLUE_DIM;                                    /* Red color dimmer value */
byte EEPROM_READ;                                 /* Testread from EEPROM */
byte UDP_PACKETS;                                 /* UDP packets */
byte SLAVE_1_MODE;                                /* Slave 1 off/auto mode (0 = off, 1 = auto) */
byte SLAVE_2_MODE;                                /* Slave 2 off/auto mode (0 = off, 1 = auto) */
byte SLAVE_3_MODE;                                /* Slave 3 off/auto mode (0 = off, 1 = auto) */
byte STORE_RED_COLOR;                             /* Red color value of selected storage position */
byte STORE_GREEN_COLOR;                           /* Green color value of selected storage position */
byte STORE_BLUE_COLOR;                            /* Blue color value of selected storage position */
byte STORE_DIM_VALUE;                             /* Dimming value of selected storage position */
int SELECT;                                       /* SELECT value */
uint16_t LOOP_COUNTER;                            /* LOOP_COUNTER */
char TIME_STRING[20] = {0};                       /* Time string e.g. 12:00 */
char DATE_STRING[20] = {0};                       /* Date string e.g. 11.12.2019 */
String RECEIVE_BUFFER = "";                       /* String to receive data */
String TEMP_BUFFER = "";                          /* String for string manipulation */


/****************************************************************************************/
/* Setup procedure                                                                      */
/****************************************************************************************/
void setup() 
{  
  int SERIAL_AVAILABLE;                           /* Wait for serial available */  
  char RECEIVE_CHAR;                              /* Received char */

  TEMP_BUFFER.reserve(TEMP_BUFFER_SIZE);          /* Reserve byte buffer for packet buffer string */ 
  RECEIVE_BUFFER.reserve (RECEIVE_BUFFER_SIZE);   /* Reserve byte buffer for receive buffer */
  
  /* Initialize FLASH size */
  EEPROM.begin(EEPROM_SIZE);                      /* Initialize FLASH with predefined size */

  /* Initialise pins */
  pinMode(DMX_DIRECTION_PIN, OUTPUT);             /* Configure DMX direction pin as output */
  digitalWrite(DMX_DIRECTION_PIN, HIGH);          /* Set DMX direction pin to high = Transmit */
  pinMode(DMX_SERIAL_OUTPUT_PIN, OUTPUT);         /* Configure DMX transmit pin as output */
  pinMode(WD_TRIGGER,INPUT);                      /* Define WATCHDOG TRIGGER as intput */
  digitalWrite(WD_TRIGGER,HIGH);                  /* Write WATCHDOG TRIGGER for Pullup */
  pinMode(WD_TRIGGER,OUTPUT);                     /* Define WATCHDOG TRIGGER as output */
  digitalWrite(WD_TRIGGER,HIGH);                  /* Switch WATCHDOG TRIGGER high */
  WD_FLAG = true;                                 /* Set WD flag true = High */

  /* Init serial interface for debug information */
  Serial.begin(115200);                           /* Start serial interface for terminal communication at 115000 Baud */
  /* Display welcome message with initialisation information */
  Serial.println("\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n");
  Serial.println ("*********************************************************");
  Serial.println ("* NEO Pixel RGB Controller for ESP32 for lounge lamp    *");
  Serial.println ("*                                                       *");
  Serial.print   ("* V");
  Serial.print   (SW_VERSION);
  Serial.println ("                                      @PT 2020    *");   
  Serial.println ("*********************************************************\n"); 

  #ifdef _DEBUG_
    Serial.println("\n\n\n"); 
    Serial.println ("Debug is enabled...");
    Serial.println(); 
  #else
    Serial.println("\n\n\n"); 
    Serial.println ("Debug is disabled...");
    Serial.println ("Remove commented out line #define _DEBUG_ to enable!");
    Serial.println();     
  # endif  

  /* Read init parameters from FLASH */
  if (EEPROM.read(0)==255)                        /* FLASH never written before? --> First intialisation */
  {
    SUB_INIT_FLASH ();                            /* Call subroutine to initialize FLASH with default values */
  }

  /* Read out FLASH values */
  #ifdef _DEBUG_
    SUB_DISPLAY_FLASH ();                         /* Dump all read FLASH values */
  # endif 

  #ifdef _DEBUG_
    Serial.println();
    Serial.println ("Step 1: Init DMX interface and switch channels off: Done");
    Serial.println();
  # endif

  /* Step 1: Initializes DMX interface */
  ESP32DMX.startOutput(DMX_SERIAL_OUTPUT_PIN);    /* Start DMX transmission */
  SUB_DMX_SEND (0,0,0);                           /* Configure the first three DMX channels with 0 = Off */ 

  SLAVE_1_MODE = 1;                               /* Set slave 1 mode to 1 to enable local display */
  SLAVE_2_MODE = 0;                               /* Set slave 2 mode to 0 to disable local display */
  SLAVE_3_MODE = 0;                               /* Set slave 3 mode to 0 to disable local display */

  /* Step 2: Initialize SPIFFS */
  #ifdef _DEBUG_
    Serial.println ("Step 2: Initialise file access and read Init.txt...");
    Serial.print("  Initialize file access: ");
  # endif
  if(!SPIFFS.begin(true))                         /* Initialisation error ? */
  {  
    #ifdef _DEBUG_                                          
      Serial.println("Failed!");
    # endif
    do                                            /* Endless loop until manual reset */
    {
      SUB_DMX_SEND (255, 0, 0);                   /* Switch all pixels to red color */
      SUB_TRIGGER_WATCHDOG ();                    /* Trigger watchdog */
      delay (500);                                /* 500ms delay */
      SUB_DMX_SEND (0, 0, 0);                     /* Switch off all pixels */
      SUB_TRIGGER_WATCHDOG ();                    /* Trigger watchdog */
      delay (500);                                /* 500ms delay */      
    }
    while (1);  
  }
  else                                            /* No error ? */
  { 
    #ifdef _DEBUG_                                            
      Serial.println("Done");                     /* Display success message */
    # endif  
    SUB_READ_INIT_FILE();                         /* Call subroutine to read init file */
    SUB_DMX_SEND (255, 255, 0);                   /* Switch all pixels to yellow color */
    delay (2000);                                 /* 2 seconds delay */
    SUB_DMX_SEND (0, 0, 0);                       /* Switch off all pixels */
    delay (2000);
  }

  /* Clear incoming serial buffer */
  while( Serial.available())
  { 
    RECEIVE_CHAR = Serial.read();                 /* Read char if available */
  }  
  #ifdef _DEBUG_ 
    Serial.println("\n\n\n"); 
  # endif
  Serial.println(">> Press enter to change LAN settings within 5 seconds ...");
  SERIAL_AVAILABLE = 0;                           /* Serial available */
  /* Wait for new char received or timeout */
  while ((Serial.available() <= 0) && (SERIAL_AVAILABLE < 500))
  {                                               
    delay (10);
    SERIAL_AVAILABLE++;                           /* Increase serial available counter */
  }
  if (SERIAL_AVAILABLE < 500)                     /* Enter configuration mode */
  {
    SUB_CONFIG_SETTING();                         /* Call subroutine to configure settings */
  }

  /* Step 3: Connect to LAN */
  #ifdef _DEBUG_
    Serial.println();
    Serial.print("Step 3: ");
  #else
    Serial.println();
    Serial.print("Booting up system ...");
  # endif  
  ETH_CONNECT = false;                            /* Set Ethernet connect to false */
  WiFi.onEvent(SUB_ETH_EVENT);                    /* Define EThernet event handler */
  ETH.begin();                                    /* Start ethernet connection */
  SUB_ETH_CONNECT();                              /* Call sub routine to connect to Ethernet */  
  SUB_DMX_SEND (255, 0, 255);                     /* Switch all pixels to purple color */
  delay (2000);                                   /* 2 seconds delay */
  SUB_DMX_SEND (0, 0, 0);                         /* Switch off all pixels */
  delay (2000); 

  /* Check normal boot or LAN bootloader */
  if (EEPROM.read(60) == 0)                       /* Normal boot ? */
  {    
    /* Step 4: Receive NTP time */
    #ifdef _DEBUG_
      Serial.println();
    # endif  
    NTP_TIME_ZONE = EEPROM.read(63);              /* Read value for NTP time zone out of FLASH */
    /* Determine automatic summertime switching */
    if (EEPROM.read(64) == 0)                     /* No automatic summertime switching ? */
    {
      NTP_SUMMERTIME = false;                     /* Set NTP summertime switching flag false */
    }
    else                                          /* Automatic summertime switching ? */
    {
      NTP_SUMMERTIME = true;                      /* Set NTP summertime switching flag true */
    }
     
    /* Calculated from NTP_TIME_ZONE, e.g. Austia = UTC + 1 hour = 3600 seconds */
    GMT_OFFSET_SEC = NTP_TIME_ZONE_VALUE[NTP_TIME_ZONE]*3600; 
    LOOP_COUNTER = 1;                             /* Set loop counter to 1 */
    do                                            /* Repeat until NTP Sync or three tries reached */
    {
      #ifdef _DEBUG_
        Serial.print("Step 4: Trial ");  
        Serial.print (LOOP_COUNTER);
        Serial.print (": ");
      #endif
      delay (500);                                /* Delay betwen NTP sync trials */  
      SUB_NTP_RECEIVE();                          /* Call subroutine to receive NTP time */ 
      LOOP_COUNTER ++;                            /* Increase loop counter */
    }
    while ((NTP_CONNECT == false) && (LOOP_COUNTER < 4));
    
    if (NTP_CONNECT == false)                     /* Error receiving NTP info ? --> Hardware reset */    
    {
      SUB_DMX_SEND (255, 0, 0);                   /* Switch all pixels to red color */
      delay (2000);                               /* 2 seconds delay */
      SUB_DMX_SEND (0, 0, 0);                     /* Switch off all pixels */
      delay (2000);                               /* 2 seconds delay */
      SUB_WAIT_FOR_WATCHDOG ();                   /* Wait for watchdog reset */
    }
    else                                          /* No error receiving NTP ? */
    {
      SUB_DMX_SEND (0, 255, 255);                 /* Switch all pixels to cyan color */
      delay (2000);                               /* 2 seconds delay */
      SUB_DMX_SEND (0, 0, 0);                     /* Switch off all pixels */
      delay (2000);                               /* 2 seconds delay */    
    }
    
    /* Step 5: Define routes and start HTTP and UDP server */
    #ifdef _DEBUG_
      Serial.println();
      Serial.println("Step 5: Starting server and routes: Done"); 
      Serial.println("  HTTP server started at port 80");
    # endif 
    SUB_DEFINE_ROUTES ();                         /* Web.ino: Call subroutine to define routes */
    server.begin();                               /* Start HTTP webserver */
    /* Start listen to UDP port 6554 */
    if (udpclient.begin (UDP_PORT)==1)            /* Start to listen to UDP port 6454 OK ? */
    {
      #ifdef _DEBUG_
        Serial.print ("  UDP configured on port ");
        Serial.print (UDP_PORT);
        Serial.println (": Done");
      # endif
      SUB_DMX_SEND (0, 255, 0);                   /* Switch all pixels to green color */
      delay (2000);                               /* 2 seconds delay */
      SUB_DMX_SEND (0, 0, 0);                     /* Switch off all pixels */
    }
    else                                          /* Start to listen to UDP port not OK ? */
    {
      #ifdef _DEBUG_
        Serial.print ("  UDP configured on port ");
        Serial.print (UDP_PORT);
        Serial.println (": Failed! Resetting ESP device: "); 
      # endif  
      SUB_DMX_SEND (255, 0, 0);                   /* Switch all pixels to red color */
      delay (2000);                               /* 2 seconds delay */
      SUB_DMX_SEND (0, 0, 0);                     /* Switch off all pixels */
      delay (2000); 
      SUB_WAIT_FOR_WATCHDOG ();                   /* Wait for watchdog reset */ 
    }

    /* Init variables */  
    POWER_OFF = true;                             /* Keep device switched off */
    POWER_WEB_OFF = true;                         /* Web display switched off */
    SELECT = 255;                                 /* Delete key buffer */ 

    /* Update clock and status informations for first time */ 
    getLocalTime(&TIMEINFO);                      /* Get actual time */
    ACT_MIN         = TIMEINFO.tm_min;            /* Extract actual minute */ 
    ACT_HOUR        = TIMEINFO.tm_hour;           /* Extract actual hour */ 
    ACT_DAY         = TIMEINFO.tm_mday;           /* Extract actual day */
    ACT_MONTH       = TIMEINFO.tm_mon + 1;        /* Extract actual month */
    ACT_YEAR_SHORT  = TIMEINFO.tm_year - 100;     /* Extrat actual year 2 digit */
    /* Generate time string */
    memset (TIME_STRING, 0 , sizeof(TIME_STRING));
    strftime(TIME_STRING,6, "%H:%M", &TIMEINFO);
    /* Generate date string */
    memset (DATE_STRING, 0 , sizeof(DATE_STRING));
    strftime(DATE_STRING,11, "%d.%m.20%y", &TIMEINFO); 
    /* Initialise actual on time hour and minute */
    ON_TIME_HOUR = SUN_DOWN_HOUR[ACT_MONTH-1];
    ON_TIME_MIN = SUN_DOWN_MIN[ACT_MONTH-1]; 
    #ifdef _DEBUG_
      Serial.println ();
      Serial.print ("On time: ");
      Serial.print (ON_TIME_HOUR);
      if (ON_TIME_MIN < 10)                       /* Display leading zero */
      {
        Serial.print (":0");
      }
      else
      {
        Serial.print (":");
      }
      Serial.print (ON_TIME_MIN);  
      Serial.print (" set for actual month ");
      Serial.println (ACT_MONTH); 
      Serial.println ("Off time: 22:00");
      Serial.println ();
      Serial.println ("Configuration finished...");
      Serial.println ();
    # endif
    ACT_PROGRAM    = EEPROM.read (0);             /* Read actual program out of FLASH */
    COLOR_DIM_STEP = EEPROM.read (1);             /* Read actual color dimming value out of FLASH */
    PROG_DIM_STEP  = EEPROM.read (2);             /* Read actual program dimming value out of FLASH */
    PROG_SPEED     = EEPROM.read (3);             /* Read program speed out of FLASH */
    /* Initialize actual slave 1 automatic mode (0=Off, 1=Auto) */
    if (EEPROM.read(5) == 0)                      /* Read slave 1 mode */
    {
      SLAVE_1_MODE = 0;                           /* Set slave 1 mode to 0 */
    }
    else
    {
      SLAVE_1_MODE = 1;                           /* Set slave 1 mode to 1 */
    }   
    /* Initialize actual slave 2 automatic mode (0=Off, 1=Auto) */
    if (EEPROM.read(7) == 0)                      /* Read slave 2 mode */
    {
      SLAVE_2_MODE = 0;                           /* Set slave 2 mode to 0 */
    }
    else
    {
      SLAVE_2_MODE = 1;                           /* Set slave 2 mode to 1 */
    }  
    /* Initialize actual slave 3 automatic mode (0=Off, 1=Auto) */
    if (EEPROM.read(9) == 0)                      /* Read slave 3 mode */
    {
      SLAVE_3_MODE = 0;                           /* Set slave 3 mode to 0 */
    }
    else
    {
      SLAVE_3_MODE = 1;                           /* Set slave 3 mode to 1 */
    } 
    SLAVE_1_CHANGED = false;                      /* Set slave 1 mode changed via web interface false = No change */
    SLAVE_2_CHANGED = false;                      /* Set slave 2 mode changed via web interface false = No change */
    SLAVE_3_CHANGED = false;                      /* Set slave 3 mode changed via web interface false = No change */ 
    TEST_COLOR      = false;                      /* Set test color flag false */
    NTP_UPDATE_HOUR     = EEPROM.read(61);        /* Read value for NTP update hour out of FLASH */
    NTP_UPDATE_MIN      = EEPROM.read(62);        /* Read value for NTP update minute out of FLASH */
    
    SUB_TRIGGER_WATCHDOG ();                      /* Call subroutine to trigger watchdog for the first time */
  }
  else                                            /* LAN bootloader ? */
  {
    SUB_LAN_BOOTLOADER ();                        /* Call subroutine for LAN bootloader */
  }
}


/****************************************************************************************/
/* Main Loop                                                                            */
/****************************************************************************************/
void loop() 
{
  getLocalTime(&TIMEINFO);                        /* Get actual time */
  NEW_SEC = TIMEINFO.tm_sec;                      /* Extract actual second */ 

  SUB_TRIGGER_WATCHDOG ();                        /* Call subroutine to trigger watchdog at central location in main loop */

  /****************************************************************/
  /* Check to enter bootloader                                    */
  /****************************************************************/ 
  if (BOOTLOADER_FLAG == true)                    /* Bootloader flag true ? */
  {
    EEPROM.write(60, 1);                          /* Store boot mode */
    EEPROM.commit();                              /* Write to FLASH memory */ 
    #ifdef _DEBUG_
      Serial.println ();
      Serial.println ("Wait for WD reset to enter bootloader ...");
    #endif

    SUB_WAIT_FOR_WATCHDOG ();                     /* Call subroutine to wait for watchdog */    
  }

  /****************************************************************/
  /* Check to reset device                                        */
  /****************************************************************/ 
  if (RESET_FLAG == true)                         /* Reset flag true ? */
  {
    #ifdef _DEBUG_
      Serial.println ();
      Serial.println ("Wait for WD reset to restart ...");
    #endif
    SUB_WRITE_INIT_FILE ();                       /* Call subroutine to save values in Init.txt */
    SUB_WAIT_FOR_WATCHDOG ();                     /* Call subroutine to wait for watchdog */      
  }

  /****************************************************************/
  /* Check to test color                                          */
  /****************************************************************/ 
  if (TEST_COLOR == true)                         /* Test color flag true ? */
  {  
    #ifdef _DEBUG_
      Serial.println ();
      Serial.println ("Set test color ...");
    #endif
    SUB_TEST_VALUE();                             /* Call subroutine to set test value */
    TEST_COLOR = false;                           /* Set test color false */
  }
 
  /****************************************************************/
  /* Check to adjust slave mode                                   */
  /****************************************************************/ 
  if (SLAVE_1_CHANGED == true)                    /* Slave 1 mode changed? */
  {
    SLAVE_1_CHANGED = false;                      /* Set slave 1 mode changed via web interface false = No change */
    if (POWER_OFF == false)                       /* Power on state ? */
    {
      /* Program for single color? */
      if ((ACT_PROGRAM >=1) && (ACT_PROGRAM <= 13))     
      {
        SELECT = ACT_PROGRAM;                     /* Override actual program */
        SUB_MAIN_COLOR();                         /* Call subroutine for single color */
      }
      /* Program for stored values? */
      if ((ACT_PROGRAM >= 15) && (ACT_PROGRAM <= 19))         
      {
        SELECT = ACT_PROGRAM;                     /* Override actual program */
        SUB_RCL_STORAGE();                        /* Call subroutine for single storage */   
      }       
    }
  }
  if (SLAVE_2_CHANGED == true)                    /* Slave 2 mode changed? */
  {
    SLAVE_2_CHANGED = false;                      /* Set slave 2 mode changed via web interface false = No change */
    if (POWER_OFF == false)                       /* Power on state ? */
    {
      /* Program for single color? */
      if ((ACT_PROGRAM >=1) && (ACT_PROGRAM <= 13))     
      {
        SELECT = ACT_PROGRAM;                     /* Override actual program */
        SUB_MAIN_COLOR();                         /* Call subroutine for single color */
      }
      /* Program for stored values? */
      if ((ACT_PROGRAM >= 15) && (ACT_PROGRAM <= 19))         
      {
        SELECT = ACT_PROGRAM;                     /* Override actual program */
        SUB_RCL_STORAGE();                        /* Call subroutine for single storage */   
      }       
    }
  }

  if (SLAVE_3_CHANGED == true)                    /* Slave 3 mode changed? */
  {
    SLAVE_3_CHANGED = false;                      /* Set slave 2 mode changed via web interface false = No change */
    if (POWER_OFF == false)                       /* Power on state ? */
    {
      /* Program for single color? */
      if ((ACT_PROGRAM >=1) && (ACT_PROGRAM <= 13))     
      {
        SELECT = ACT_PROGRAM;                     /* Override actual program */
        SUB_MAIN_COLOR();                         /* Call subroutine for single color */
      }
      /* Program for stored values? */
      if ((ACT_PROGRAM >= 15) && (ACT_PROGRAM <= 19))         
      {
        SELECT = ACT_PROGRAM;                     /* Override actual program */
        SUB_RCL_STORAGE();                        /* Call subroutine for single storage */   
      }       
    }
  }

  /****************************************************************/
  /* Check for UDP client requests                                */
  /****************************************************************/ 
  UDP_PACKETS = udpclient.parsePacket ();  
  if (UDP_PACKETS > 0)                            /* UDP packets available ? */
  { 
    SUB_PROCESS_UDP_REQUEST();                    /* Call subroutine to handle UDP client request */
  }   

  /* Check new second */
  if (ACT_SEC != NEW_SEC)                         /* New second ? */
  {
    ACT_SEC = NEW_SEC;                            /* Save second */
    if (ACT_SEC == 0)                             /* Update clock display every minute --> Second = 0? */
    {
      ACT_MIN         = TIMEINFO.tm_min;          /* Extract actual minute */ 
      ACT_HOUR        = TIMEINFO.tm_hour;         /* Extract actual hour */ 
      ACT_DAY         = TIMEINFO.tm_mday;         /* Extract actual day */
      ACT_MONTH       = TIMEINFO.tm_mon + 1;      /* Extract actual month */
      ACT_YEAR_SHORT  = TIMEINFO.tm_year - 100;   /* Extrat actual year 2 digit */
      /* Generate time string */
      memset (TIME_STRING, 0 , sizeof(TIME_STRING));
      strftime(TIME_STRING,6, "%H:%M", &TIMEINFO);
      /* Generate date string */
      memset (DATE_STRING, 0 , sizeof(DATE_STRING));
      strftime(DATE_STRING,11, "%d.%m.20%y", &TIMEINFO);      
      #ifdef _DEBUG_
        SUB_CLOCK_PRINT ();                       /* Call sub routine to send clock and date via serial interface to terminal */
      # endif   

      /****************************************************************/
      /* Check for NTP update at specified time                       */
      /****************************************************************/
      if ((ACT_HOUR == NTP_UPDATE_HOUR) && (ACT_MIN == NTP_UPDATE_MIN))
      {
        LOOP_COUNTER = 1;                         /* Set loop counter to 1 */
        do                                        /* Repeat until NTP Sync or three tries reached */
        {
          #ifdef _DEBUG_
            Serial.print("NTP Sync: Trial ");  
            Serial.print (LOOP_COUNTER);
            Serial.print (": ");
            #endif
          delay (500);                            /* Delay betwen NTP sync trials */  
          SUB_NTP_RECEIVE();                      /* Call subroutine to receive NTP time */ 
          LOOP_COUNTER ++;                        /* Increase loop counter */
        }
        while ((NTP_CONNECT == false) && (LOOP_COUNTER < 4));
        #ifdef _DEBUG_
          Serial.println();
        # endif  

        if (NTP_CONNECT == false)                 /* Error receiving NTP info ? --> Hardware reset */
        {
          RESET_FLAG = true;                      /* Set reset flag true */
        }
        else                                      /* No error receiving NTP info? */
        {
          ON_TIME_HOUR = SUN_DOWN_HOUR[ACT_MONTH-1];
          ON_TIME_MIN = SUN_DOWN_MIN[ACT_MONTH-1];
        }
      }

      /****************************************************************/
      /* Check for Ethernet connection every 30 minutes               */
      /****************************************************************/
      if ((ACT_MIN == 4) || (ACT_MIN == 34))      /* Check connection LAN Status now ? */
      {
        SUB_CHECK_ETH_CONNECT();                  /* Call subroutine to check if Ethernet is still connected */
      }

      /****************************************************************/
      /* Check for switch on based on hour information in Auto Mode   */
      /****************************************************************/
      if ((ACT_HOUR == ON_TIME_HOUR) && (ACT_MIN == ON_TIME_MIN))
      {
        TEST_POS_1 = EEPROM.read (4);             /* Read mode bit from FLASH */
        if (TEST_POS_1 != 0)                      /* Not in off mode ? */
        {
          SELECT=129;                             /* Switch on */
        }
      }

      /****************************************************************/
      /* Check for switch off                                         */
      /****************************************************************/
      if ((ACT_HOUR == 22) && (ACT_MIN == 00))
      {
        SELECT=128;                               /* Switch off */
      }
    }
  }

  /****************************************************************/
  /* Power Off Selected                                           */
  /****************************************************************/
  if ((SELECT==128) && (POWER_OFF == false))      /* Device in power on state? */
  {
    SUB_POWER_OFF();                              /* Call subroutine for power off */
  }

  /****************************************************************/
  /* Power On Selected                                            */
  /****************************************************************/
  if ((SELECT==129) && (POWER_OFF == true))       /* Device in power off state? */
  {
    ACT_PROGRAM    = EEPROM.read (0);             /* Read actual program out of FLASH */
    COLOR_DIM_STEP = EEPROM.read (1);             /* Read actual color dimming value out of FLASH */
    PROG_DIM_STEP  = EEPROM.read (2);             /* Read actual program dimming value out of FLASH */
    PROG_SPEED     = EEPROM.read (3);             /* Read program speed out of FLASH */
    SELECT = ACT_PROGRAM;                         /* Determine correct SELECT value */
    #ifdef _DEBUG_
      Serial.println ("");
      Serial.println ("Switching power on...");
      Serial.print ("Actual program: ");
      Serial.println (ACT_PROGRAM);
    # endif  
    POWER_OFF = false;                            /* Set power off flag false */
    POWER_WEB_OFF = false;                        /* Set power off flag for web false */
    
    if ((SELECT >=1) && (SELECT <= 13))           /* Program for single color? */
    {
      SUB_MAIN_COLOR();                           /* Call subroutine for single color */
    }
    
    if ((SELECT >= 15) && (SELECT <= 19))         /* Program for stored values? */
    {
      SUB_RCL_STORAGE();                          /* Call subroutine for single storage */   
    }        
  
    if ((SELECT >= 20) && (SELECT <= 29))         /* Program for program sequence? */
    {
      SUB_PROG_SEL ();                            /* Call subroutine for animated programs */
    }
  }

  /****************************************************************/
  /* Program 1-13 (single color 1-13) selected                    */
  /****************************************************************/
  if ((SELECT >=1) && (SELECT <= 13))
  {
    if (POWER_OFF == true)                        /* Switched on from power off state */
    {
      POWER_OFF = false;                          /* Set power off false */
      POWER_WEB_OFF = false;                      /* Set power off false for web */
      COLOR_DIM_STEP = EEPROM.read (1);           /* Read actual color dimming value out of FLASH */
    }
    if (UDP_COLOR == true)                        /* Unfinished color savings ? */
    {
      UDP_COLOR = false;                          /* Reset UDP_COLOR flag */ 
      COLOR_DIM_STEP = EEPROM.read (1);           /* Restore actual color dimming value out of FLASH */ 
    }
    SUB_MAIN_COLOR ();                            /* Call subroutine for single color with restore */
  } 
  
  /****************************************************************/
  /* Program 15-19 (Recall 1-5) selected                          */
  /****************************************************************/
  if ((SELECT >=15) && (SELECT <= 19))
  {
    if (POWER_OFF == true)                        /* Switched on from power off state */
    {
      POWER_OFF = false;                          /* Set power off false */
      POWER_WEB_OFF = false;                      /* Set power off false for web */
    } 
    UDP_COLOR = false;                            /* Reset UDP_COLOR flag */    
    SUB_RCL_STORAGE();                            /* Call subroutine for single storage */  
  } 
  
  /****************************************************************/
  /* Program 20-29 (Program 1-5) selected                         */
  /****************************************************************/
  if ((SELECT >=20) && (SELECT <= 29))
  {
    if (POWER_OFF == true)                        /* Switched on from power off state */
    {
      POWER_OFF = false;                          /* Set power off false */
      POWER_WEB_OFF = false;                      /* Set power off false for web */
      PROG_DIM_STEP = EEPROM.read (2);            /* Read actual color dimming value out of FLASH */
      PROG_SPEED    = EEPROM.read (3);            /* Read program speed out of FLASH */
    }
    SUB_PROG_SEL ();                              /* Call subroutine for animated programs */  
    UDP_COLOR = false;                            /* Reset UDP_COLOR flag */
  }   
      
  /****************************************************************/
  /* Program 32-40 (Color dimming factor 1-8) selected            */
  /****************************************************************/   
  if ((SELECT >= 32)&& (SELECT <= 40) && (POWER_OFF == false))
  {
    COLOR_DIM_STEP = SELECT - 32;                 /* Determine dimming factor */
    #ifdef _DEBUG_
      Serial.println ("");
      Serial.print ("Colour dimming factor ");
      Serial.println (COLOR_DIM_STEP);
      Serial.println (" selected ...");
    # endif      

    if (UDP_COLOR == false)                       /* No unsaved color ? --> Save */
    {    
      /* Program for single color? */
      if ((ACT_PROGRAM >=1) && (ACT_PROGRAM <= 13))           
      {
        SELECT = ACT_PROGRAM;                     /* Copy sctual program to SELECT variable */
        SUB_MAIN_COLOR();                         /* Call subroutine for single color */
      }
      /* Program for stored values? */
      if ((ACT_PROGRAM >= 15) && (ACT_PROGRAM <= 19))         
      {
        /* Calculate red color value */
        RED_COLOR = MAX_RED_COLOR - COLOR_DIM_STEP * RED_DIM; 
        /* Calculate green color value */
        GREEN_COLOR = MAX_GREEN_COLOR - COLOR_DIM_STEP * GREEN_DIM; 
        /* Calculate blue color value */
        BLUE_COLOR = MAX_BLUE_COLOR - COLOR_DIM_STEP * BLUE_DIM;  
        /* Call subroutine send via DMX interface */            
        SUB_DMX_SEND (RED_COLOR, GREEN_COLOR, BLUE_COLOR);       
      } 
    }
    else                                          /* Unsaved color ? */
    {
      /* Calculate red color value */
      RED_COLOR = MAX_RED_COLOR - COLOR_DIM_STEP * RED_DIM;
      /* Calculate green color value */
     GREEN_COLOR = MAX_GREEN_COLOR - COLOR_DIM_STEP * GREEN_DIM; 
     /* Calculate blue color value */ 
     BLUE_COLOR = MAX_BLUE_COLOR - COLOR_DIM_STEP * BLUE_DIM;  
     /* Call subroutine send via DMX interface */          
     SUB_DMX_SEND (RED_COLOR, GREEN_COLOR, BLUE_COLOR);        
    }   
    delay (100);                                  /* 100ms delay */
    SELECT = 255;                                 /* Set command to 255 */   
  }    
  
  /****************************************************************/
  /* Program 136-140 (Store 1-5) selected                         */
  /****************************************************************/
  if ((SELECT >=136) && (SELECT <= 140) && (POWER_OFF == false))
  {
    SUB_SET_STORAGE();                            /* Call subroutine for set storage */   
  }    
  
  /****************************************************************/
  /* Program 143 restore memory default selected                  */
  /****************************************************************/
  if (SELECT == 143)                              /* Restore color storage positions 1-5 */
  { 
    #ifdef _DEBUG_
      Serial.println ("");
      Serial.println ("Restoring storage positions 1-5 ... ");
    # endif      
    EEPROM.write(10, 144);                        /* Storage position 1: Red color value */ 
    EEPROM.write(11, 72);                         /* Storage position 1: Green color value */ 
    EEPROM.write(12, 18);                         /* Storage position 1: Blue color value */ 
    EEPROM.write(13,0);                           /* Storage position 1: Dimming factor = 0 */  
    EEPROM.write(20, 252);                        /* Storage position 2: Red color value */
    EEPROM.write(21, 45);                         /* Storage position 2: Green color value */
    EEPROM.write(22, 0);                          /* Storage position 2: Blue color value */
    EEPROM.write(23,0);                           /* Storage position 2: Dimming factor = 0 */ 
    EEPROM.write(30, 72);                         /* Storage position 3: Red color value */
    EEPROM.write(31, 45);                         /* Storage position 3: Green color value */
    EEPROM.write(32, 0);                          /* Storage position 3: Blue color value */
    EEPROM.write(33,0);                           /* Storage position 3: Dimming factor = 0 */ 
    EEPROM.write(40, 112);                        /* Storage position 4: Red color value */
    EEPROM.write(41, 0);                          /* Storage position 4: Green color value */
    EEPROM.write(42, 56);                         /* Storage position 4: Blue color value */
    EEPROM.write(43,0);                           /* Storage position 4: Dimming factor = 0 */ 
    EEPROM.write(50, 27);                         /* Storage position 5: Red color value */
    EEPROM.write(51, 99);                         /* Storage position 5: Green color value */
    EEPROM.write(52, 252);                        /* Storage position 5: Blue color value */
    EEPROM.write(53,0);                           /* Storage position 5: Dimming factor = 0 */ 
    EEPROM.commit();                              /* Store values permanently in EEPROM */ 
    delay (100);                                  /* 100ms delay */
    SELECT = 255;                                 /* Set command to 255 */
  }
}
