/****************************************************************************************/
/* LAN.ino                                                                              */
/* Collection of subroutines for LAN and NTP                                            */
/* 1. LAN routines:                                                                     */
/* SUB_ETH_EVENT:         Subroutine Ethernet event handler                             */
/* SUB_ETH_CONNECT:       Subroutine to connect to LAN                                  */
/* SUB_CHECK_ETH_CONNECT: Subroutine to check if LAN connection is still on             */
/*                                                                                      */
/* 2. NTP routines:                                                                     */
/* SUB_NTP_RECEIVE:           Subroutine to receive NTP time                            */
/* SUB_NTP_ADJUST_SUMMERTIME: Subroutine to adjust received time to timezone            */          
/****************************************************************************************/


/****************************************************************************************/
/****************************************************************************************/
/* LAN routines                                                                         */
/****************************************************************************************/
/****************************************************************************************/


/****************************************************************************************/
/* Subroutine Ethernet event handler                                                    */
/****************************************************************************************/
void SUB_ETH_EVENT(WiFiEvent_t ETH_EVENT)
{
  switch (ETH_EVENT) 
  {   
    case SYSTEM_EVENT_ETH_START:                  /* Ethernet start event */
      /* Set fixed IP address */
      ETH.config(FIXED_IP, GATEWAY, SUBNET, DNS_1, DNS_2);
      ETH.setHostname(HOSTNAME);                  /* Set hostname */
      break;
    case SYSTEM_EVENT_ETH_CONNECTED:              /* Ethernet connected event */
      break;
    case SYSTEM_EVENT_ETH_GOT_IP:                 /* Ethernet IP address received */
      ETH_CONNECT = true;                         /* Set Ethernet connect flag true */
      break;
    case SYSTEM_EVENT_ETH_DISCONNECTED:           /* Ethernet disconnected ? */
      ETH_CONNECT = false;                        /* Set Ethernet connect flag false */   
      break;
    case SYSTEM_EVENT_ETH_STOP:                   /* Ethernet stopped ? */
      ETH_CONNECT = false;                        /* Set Ethernet connect flag false */   
      break;
    default:                                      /* Default situation = None */
      break;
  }
}


/****************************************************************************************/
/* Subroutine to connect to LAN                                                         */
/****************************************************************************************/
void SUB_ETH_CONNECT (void)
{
  boolean  PING_STATUS;                           /* True = Responding to ping, False = No response to ping */
  
  #ifdef _DEBUG_
    Serial.println("Starting to connect to LAN ..."); 
  # endif 
  do                                              /* Repeat until LAN is connected */
  {
    delay (10);                                   /* 10ms delay */
  }
  while (ETH_CONNECT == false);

  #ifdef _DEBUG_
    Serial.print ("  Ping server for connectivity check: ");
    Serial.print (GATEWAY);
    Serial.print(": "); 
  # endif
  PING_STATUS = Ping.ping(GATEWAY, 4);            /* Ping to IP adress */
  if (PING_STATUS == true)                        /* Sucessful ping? */
  {
    #ifdef _DEBUG_
      Serial.println("Success");
      /* Display LAN information */
      Serial.print("    IP address: ");
      Serial.println(ETH.localIP());
      Serial.print("    ESP Mac Address: ");
      Serial.println(ETH.macAddress());
      Serial.print("    Subnet Mask: ");
      Serial.println(ETH.subnetMask());
      Serial.print("    Gateway IP: ");
      Serial.println(ETH.gatewayIP()); 
      Serial.print("    Hostname: ");
      Serial.println(ETH.getHostname());         
      Serial.print("    Link speed: ");
      Serial.print(ETH.linkSpeed());
      Serial.println(" Mbps");
      Serial.print("    Link mode: ");   
      if (ETH.fullDuplex()) 
      {
        Serial.println (" FULL_DUPLEX");
      }
      else
      {
        Serial.println (" HALVE_DUPLEX");
      }                        
      Serial.print("    Use this URL: http://");
      Serial.print(ETH.localIP());
      Serial.println("/main");          
    # endif 
  }
  else                                            /* Unsucessful ping? */
  {
    #ifdef _DEBUG_
      Serial.println("Failed");
    # endif
    SUB_WAIT_FOR_WATCHDOG ();                     /* Wait for watchdog reset */
  }
}


/****************************************************************************************/
/* Subroutine to check if LAN connection is still on                                    */
/****************************************************************************************/
void SUB_CHECK_ETH_CONNECT (void)
{
  boolean  PING_STATUS;                           /* True = Responding to ping, False = No response to ping */

  #ifdef _DEBUG_
    Serial.print ("Checking LAN connection: ");
  # endif 

  PING_STATUS = Ping.ping(GATEWAY, 4);            /* Ping to IP adress */
  if (PING_STATUS == true)                        /* Sucessful ping? */
  {
    #ifdef _DEBUG_
      Serial.println("Ok");
    # endif
  }
  else                                          /* Unsucessful ping? */
  {
    #ifdef _DEBUG_
      Serial.println("Error");
    # endif
  }
  #ifdef _DEBUG_
    Serial.println();
  # endif
}


/****************************************************************************************/
/****************************************************************************************/
/* NTP routines                                                                         */
/****************************************************************************************/
/****************************************************************************************/


/****************************************************************************************/
/* NTP-Clock: Subroutine to receive NTP time                                            */
/*                                                                                      */
/* %a Abbreviated weekday name                                                          */
/* %A Full weekday name                                                                 */
/* %b Abbreviated month name                                                            */
/* %B Full month name                                                                   */
/* %c Date and time representation for your locale                                      */
/* %d Day of month as a decimal number (01-31)                                          */
/* %H Hour in 24-hour format (00-23)                                                    */
/* %I Hour in 12-hour format (01-12)                                                    */
/* %j Day of year as decimal number (001-366)                                           */
/* %m Month as decimal number (01-12)                                                   */
/* %M Minute as decimal number (00-59)                                                  */
/* %p Current locale's A.M./P.M. indicator for 12-hour clock                            */
/* %S Second as decimal number (00-59)                                                  */
/* %U Week of year as decimal number,  Sunday as first day of week (00-51)              */
/* %w Weekday as decimal number (0-6; Sunday is 0)                                      */
/* %W Week of year as decimal number, Monday as first day of week (00-51)               */
/* %x Date representation for current locale                                            */
/* %X Time representation for current locale                                            */
/* %y Year without century, as decimal number (00-99)                                   */
/* %Y Year with century, as decimal number                                              */
/* %z %Z Time-zone name or abbreviation, (no characters if time zone is unknown)        */
/* %% Percent sign                                                                      */
/* You can include text literals (such as spaces and colons) to make a neater display   */
/* or for padding between adjoining columns.                                            */
/* You can suppress the display of leading zeroes  by using the "#" character           */
/* (%#d, %#H, %#I, %#j, %#m, %#M, %#S, %#U, %#w, %#W, %#y, %#Y)                         */
/*                                                                                      */
/* Variables of structure tm:                                                           */
/*  int tm_sec                                                                          */
/*  int tm_min                                                                          */
/*  int tm_hour                                                                         */
/*  int tm_mday                                                                         */
/*  int tm_mon                                                                          */
/*  int tm_year                                                                         */
/*  int tm_wday                                                                         */
/*  int tm_yday                                                                         */
/*  int tm_isdst                                                                        */
/****************************************************************************************/
void SUB_NTP_RECEIVE()
{
  NTP_CONNECT = false;                            /* Set NTP connect = true */

  #ifdef _DEBUG_
    Serial.println("Retrieving NTP time and date ...");
    Serial.print("  Receive first NTP Sync: ");  
  # endif   

  DAYLIGHT_OFFSET_SEC = 0;                        /* Set to winter time */
  /* Gets the actual time via NTP server */
  configTime(GMT_OFFSET_SEC, DAYLIGHT_OFFSET_SEC, NTP_SERVER_1, NTP_SERVER_2, NTP_SERVER_3);  
  if(!getLocalTime(&TIMEINFO))                    /* Reception of NTP time not OK ? */
  {
    #ifdef _DEBUG_
      Serial.println(" Failed");
    # endif
    return;                                       /* Return from subroutine */
  }
  else                                            /* Reception of NTP time OK ? */
  {
    #ifdef _DEBUG_
      Serial.println(" Done"); 
      Serial.print ("    Time: ");
      Serial.print(&TIMEINFO, "%H:%M:%S");
      Serial.print (", Date: ");
      Serial.println (&TIMEINFO, "%d.%m.%Y");
    # endif
    ACT_YEAR = TIMEINFO.tm_year +1900;            /* Extract actual year */
    ACT_MONTH = TIMEINFO.tm_mon + 1;              /* Extract actual month */
    ACT_DAY = TIMEINFO.tm_mday;                   /* Extract actual day */     
    /* Summertime detected and NTP summertime flag = true = ummertime */
    if((SUB_NTP_ADJUST_SUMMERTIME() == true) && (NTP_SUMMERTIME == true))    
    {
      DAYLIGHT_OFFSET_SEC = 3600;                 /* Set to summertime */ 
      #ifdef _DEBUG_
        Serial.println ("    Summertime detected");
        Serial.print("  Receive second NTP Sync: ");  
      # endif  
  
      /* Gets the actual time via NTP server */
      configTime(GMT_OFFSET_SEC, DAYLIGHT_OFFSET_SEC, NTP_SERVER_1, NTP_SERVER_2, NTP_SERVER_3);  
      if(!getLocalTime(&TIMEINFO))                /* Reception of NTP time not OK ? */
      {
        #ifdef _DEBUG_
          Serial.println(" Failed");
        # endif
        return;                                   /* Return from subroutine */
      }
      else                                        /* Reception of NTP time OK ? */
      {
        #ifdef _DEBUG_
          Serial.println(" Done"); 
          Serial.print ("    Time: ");
          Serial.print(&TIMEINFO, "%H:%M:%S");
          Serial.print (", Date: ");
          Serial.println (&TIMEINFO, "%d.%m.%Y");
        # endif  
        NTP_CONNECT = true;                       /* Set NTP connect = true */
        return;                                   /* Return from subroutine */ 
      } 
    }
    else                                          /* Wintertime detected or no automatic detection of summertime ? */
    {
      #ifdef _DEBUG_
        if (NTP_SUMMERTIME == false)              /* No automatic summertime detection ? */
        {
          Serial.print ("    Automatic summertie switching disabled");       
        }
        else                                      /* Automatic summertime detection ? */
        {
          Serial.print ("    Wintertime detected");
        } 
      # endif
      DAYLIGHT_OFFSET_SEC = 0;                    /* Set to wintertime */ 
      NTP_CONNECT = true;                         /* Set NTP connect = true */
      return;                                     /* Return from subroutine */ 
    }
  }
}


/****************************************************************************************/
/* NTP-Clock: Subroutine to adjust received time to timezone                            */
/* Returns tue if summertine and false if wintertime                                    */            
/****************************************************************************************/
boolean SUB_NTP_ADJUST_SUMMERTIME()
{
  int BEGIN_DST_DATE;                             /* Begin date */
  int BEGIN_DST_MONTH;                            /* Begin month */
  int END_DST_DATE;                               /* End date */
  int END_DST_MONTH;                              /* End month */

  /* Determine last Sunday of March */
  BEGIN_DST_DATE = (31 - (5* ACT_YEAR /4 + 4) % 7); 
  BEGIN_DST_MONTH = 3;                            /* Begin month is March */
  #ifdef _DEBUG_
    Serial.print ("    Actual summer time begin date: ");
    Serial.print (BEGIN_DST_DATE);
    Serial.print (".");
    Serial.println (BEGIN_DST_MONTH);  
  # endif  
  /* Determine last Sunday of October */
  END_DST_DATE = (31 - (5 * ACT_YEAR /4 + 1) % 7);   
  END_DST_MONTH = 10;                             /* Begin month is October */
  #ifdef _DEBUG_
    Serial.print ("    Actual summer time end date: ");
    Serial.print (END_DST_DATE);
    Serial.print (".");
    Serial.println (END_DST_MONTH);  
  # endif      
  
  /* Check for summer time */
  if (((ACT_MONTH > BEGIN_DST_MONTH) && (ACT_MONTH < END_DST_MONTH))
      || ((ACT_MONTH == BEGIN_DST_MONTH) && (ACT_DAY >= BEGIN_DST_DATE))
      || ((ACT_MONTH == END_DST_MONTH) && (ACT_DAY <= END_DST_DATE)))
  {
    return true;                                  /* Return true = Summer time */
  }
  else                                            /* Winter time ? */
  {
    return false;                                 /* Return true = Summer time */
  }
} 
