/****************************************************************************************/
/* Basis.ino                                                                            */
/* Collection of basis routines                                                         */
/* 1. FLASH routines:                                                                   */
/* SUB_INIT_FLASH:          Subroutine to initialize FLASH with default values          */
/*                                                                                      */
/* 2. INIT FILE routines:                                                               */
/* SUB_READ_INIT_FILE:      Subroutine to read out init file Init.txt                   */
/* SUB_WRITE_INIT_FILE:     Subroutine to save settings in init file Init.txt           */
/* SUB_CONFIG_SETTING:      Subroutine to change configuration settings via             */
/*                          console entry                                               */
/* 3. DEBUG routines:                                                                   */
/* SUB_DISPLAY_FLASH:       Subroutine to display all FLASH values                      */
/* SUB_CLOCK_PRINT:         Subroutine to send clock and date via serial interface      */
/*                          to terminal                                                 */
/* 4. WATCHDOG routines:                                                                */
/* SUB_TRIGGER_WATCHDOG:    Subroutine to trigger watchdog                              */
/* SUB_WAIT_FOR_WATCHDOG:   Subroutine wait for hardware reset via watchog reset        */
/*                                                                                      */
/* 5. BOOTLOADER routines:                                                              */
/* SUB_WLAN_BOOTLOADER:     Subroutine to initialize WLAN bootloader and wait           */
/*                          for upload                                                  */
/* 6. DMX routines:                                                                     */
/* SUB_DMX_SEND:            Subroutine send DMX channels                                */
/****************************************************************************************/


/****************************************************************************************/
/****************************************************************************************/
/* FLASH ROUTINES                                                                       */
/****************************************************************************************/
/****************************************************************************************/

/****************************************************************************************/
/* Initialize FLASH with default values                                                 */
/****************************************************************************************/
void SUB_INIT_FLASH (void)
{
  #ifdef _DEBUG_
    Serial.println ("Initialisation of FLASH memory ...");  
  # endif 
  EEPROM.write(0, 1);                             /* Storage position 0: Color 1 */ 
  EEPROM.write(1, 0);                             /* No dimming color */ 
  EEPROM.write(2, 0);                             /* No dimming program */ 
  EEPROM.write(3, 4);                             /* Program speed = 4 */   
  EEPROM.write(4, 0);                             /* Mode = 0 = Off */ 
  EEPROM.write(5, 1);                             /* Slave 1 mode = 1 = Off */ 
  EEPROM.write(7, 0);                             /* Slave 2 mode = 0 = Off */ 
  EEPROM.write(9, 0);                             /* Slave 3 mode = 0 = Off */     
  EEPROM.write(10, 144);                          /* Storage position 1: Red color value */ 
  EEPROM.write(11, 72);                           /* Storage position 1: Green color value */ 
  EEPROM.write(12, 18);                           /* Storage position 1: Blue color value */ 
  EEPROM.write(13,0);                             /* Storage position 1: Dimming factor = 0 */  
  EEPROM.write(20, 252);                          /* Storage position 2: Red color value */
  EEPROM.write(21, 45);                           /* Storage position 2: Green color value */
  EEPROM.write(22, 0);                            /* Storage position 2: Blue color value */
  EEPROM.write(23,0);                             /* Storage position 2: Dimming factor = 0 */ 
  EEPROM.write(30, 72);                           /* Storage position 3: Red color value */
  EEPROM.write(31, 45);                           /* Storage position 3: Green color value */
  EEPROM.write(32, 0);                            /* Storage position 3: Blue color value */
  EEPROM.write(33,0);                             /* Storage position 3: Dimming factor = 0 */ 
  EEPROM.write(40, 112);                          /* Storage position 4: Red color value */
  EEPROM.write(41, 0);                            /* Storage position 4: Green color value */
  EEPROM.write(42, 56);                           /* Storage position 4: Blue color value */
  EEPROM.write(43,0);                             /* Storage position 4: Dimming factor = 0 */ 
  EEPROM.write(50, 27);                           /* Storage position 5: Red color value */
  EEPROM.write(51, 99);                           /* Storage position 5: Green color value */
  EEPROM.write(52, 252);                          /* Storage position 5: Blue color value */
  EEPROM.write(53,0);                             /* Storage position 5: Dimming factor = 0 */  
  EEPROM.write(60, 0);                            /* Init storage position: REMOTE UPLOAD FLAG: 0=Normal, 1=Remote update */
  EEPROM.write(61, 4);                            /* Init storage position: NTP UPDATE HOUR: Hour = 3 */
  EEPROM.write(62, 15);                           /* Init storage position: NTP UPDATE MIN: Minute = 34 */
  EEPROM.write(63, 6);                            /* Init storage position: NTP TIME ZONE: Time zone = +1 hour */
  EEPROM.write(64, 1);                            /* Init storage position: NTP SUMMERTIME: Automatic switching = 1 */   
  EEPROM.commit();                                /* Write to FLASH memory */
}


/****************************************************************************************/
/****************************************************************************************/
/* INIT FILE ROUTINES                                                                   */
/****************************************************************************************/
/****************************************************************************************/


/****************************************************************************************/
/* Subroutine to read out init file Init.txt                                            */
/****************************************************************************************/
void SUB_READ_INIT_FILE (void)
{
  boolean BREAK_LOOP;                             /* Break loop flag: false = No break, true = Break */
  char RECEIVE_CHAR;                              /* Received byte from file reading */
  byte LINE_COUNTER;                              /* Line counter */
  byte IP_ADDRESS_COUNTER;                        /* IP adress counter */
  int TEST_INT;                                   /* Test integer */
  String IP_ADRESS_STRING;                        /* IP Adressstring */

  #ifdef _DEBUG_
    Serial.print("  Open file Init.txt: ");
  # endif 

  BREAK_LOOP = false;                             /* Set break loop false */
  File file = SPIFFS.open("/Init.txt", FILE_READ);/* Open Init file for read */
  if(!file)                                       /* File does not exist? */
  {
    #ifdef _DEBUG_
      Serial.println ("Failed! File does not exist");
    # endif
    BREAK_LOOP = true;                            /* Set break loop true */ 
  }
  
  if (BREAK_LOOP == false)                        /* No error yet */
  {
    #ifdef _DEBUG_
      Serial.println ("Done");
      Serial.println ("  Read Init.txt content: ");
    # endif  
    LINE_COUNTER = 1;                             /* Set line counter to first line */
    LOOP_COUNTER = 0;                             /* Check for maximum received bytes to prevent buffer overflows */

    /* Delete WLAN_SSID string */
    memset (WLAN_SSID, 0 , sizeof(WLAN_SSID));
    /* Delete WIFI passowrd string */
    memset (WLAN_SSID, 0 , sizeof(WLAN_PWD));
    /* Delete HOSTNAME string */
    memset (HOSTNAME, 0 , sizeof(HOSTNAME));
    /* Delete NTP server 1 string */
    memset (NTP_SERVER_1, 0 , sizeof(NTP_SERVER_1));
    /* Delete NTP server 2 string */
    memset (NTP_SERVER_2, 0 , sizeof(NTP_SERVER_2));
    /* Delete NTP server 3 string */
    memset (NTP_SERVER_3, 0 , sizeof(NTP_SERVER_3));    
       
    /* Read until end of file or error detected */
    while((file.available()) && (BREAK_LOOP == false))                    
    {
      RECEIVE_CHAR = file.read();                 /* Read char byte from file */

      if (RECEIVE_CHAR == '\r')                   /* Carrage return detected ? */
      {

        /* NTP 2 server */
        if (LINE_COUNTER == 10)                   /* Receive NTP 2 name ? */
        {
          #ifdef _DEBUG_
            Serial.print ("    NTP 2 server: ");
          # endif 
          /* Check on too long NTP 2 server name */
          if ((LOOP_COUNTER > 1) && (LOOP_COUNTER < 20))        
          {
            #ifdef _DEBUG_
              Serial.println (NTP_SERVER_2);
            # endif 
            LINE_COUNTER = 11;                    /* Set line counter to eleventh line */
            LOOP_COUNTER = 0;                     /* Reset loop counter */            
          }
          else                                    /* NTP 2 length error */
          {
             #ifdef _DEBUG_
              Serial.println ("NTP server 2 length error!");
            # endif 
            BREAK_LOOP = true;                    /* Set break loop true */
          }
        }  

        /* NTP 1 server */
        if (LINE_COUNTER == 9)                    /* Receive NTP 1 name ? */
        {
          #ifdef _DEBUG_
            Serial.print ("    NTP 1 server: ");
          # endif 
          /* Check on too long NTP 1 server name */
          if ((LOOP_COUNTER > 1) && (LOOP_COUNTER < 20))        
          {
            #ifdef _DEBUG_
              Serial.println (NTP_SERVER_1);
            # endif 
            LINE_COUNTER = 10;                    /* Set line counter to tenth line */
            LOOP_COUNTER = 0;                     /* Reset loop counter */            
          }
          else                                    /* NTP 1 length error */
          {
             #ifdef _DEBUG_
              Serial.println ("NTP server 1 length error!");
            # endif 
            BREAK_LOOP = true;                    /* Set break loop true */
          }
        }        

       /* DNS 2 IP address */
        if (LINE_COUNTER == 8)                    /* DNS 2 IP address ? */
        {
          #ifdef _DEBUG_
            Serial.print ("    DNS 2 IP-Address: ");
          # endif 
          TEST_INT = IP_ADRESS_STRING.toInt();
          /* In range ? */
          if ((TEST_INT >= 0) && (TEST_INT <= 255))
          {
            DNS_2[3] = TEST_INT;                  /* Save 3rd digit */  
            IP_ADDRESS_COUNTER = 5;               /* Set to fifth address */                          
          }
          else                                    /* Not in range ? */
          {
            LOOP_COUNTER = 50;                    /* Set loop counter out of range */
          }
          /* Check if 4 digits received and loop counter in range */
          if ((LOOP_COUNTER < 20) && (IP_ADDRESS_COUNTER == 5))        
          {
            #ifdef _DEBUG_
              Serial.println (DNS_2);
            # endif 
            LINE_COUNTER = 9;                     /* Set line counter to ninth line */
            LOOP_COUNTER = 0;                     /* Reset loop counter */
          }
          else                                    /* Error ? */  
          {
            #ifdef _DEBUG_
              Serial.println ("DNS 2 IP-Address format error!");
            # endif 
            BREAK_LOOP = true;                    /* Set break loop true */
          }
        }

       /* DNS 1 IP address */
        if (LINE_COUNTER == 7)                    /* DNS 1 IP address ? */
        {
          #ifdef _DEBUG_
            Serial.print ("    DNS 1 IP-Address: ");
          # endif 
          TEST_INT = IP_ADRESS_STRING.toInt();
          /* In range ? */
          if ((TEST_INT >= 0) && (TEST_INT <= 255))
          {
            DNS_1[3] = TEST_INT;                  /* Save 3rd digit */ 
            IP_ADDRESS_COUNTER = 5;               /* Set to fifth address */                           
          }
          else                                    /* Not in range ? */
          {
            LOOP_COUNTER = 50;                    /* Set loop counter out of range */
          }
          /* Check if 4 digits received and loop counter in range */
          if ((LOOP_COUNTER < 20) && (IP_ADDRESS_COUNTER == 5))        
          {
            #ifdef _DEBUG_
              Serial.println (DNS_1);
            # endif 
            LINE_COUNTER = 8;                     /* Set line counter to eighth line */
            IP_ADDRESS_COUNTER = 1;               /* Set to first address */
            LOOP_COUNTER = 0;                     /* Reset loop counter */
            IP_ADRESS_STRING = "";                /* Delete string for next digit */
          }
          else                                    /* Error ? */  
          {
            #ifdef _DEBUG_
              Serial.println ("DNS 1 IP-Address format error!");
            # endif 
            BREAK_LOOP = true;                    /* Set break loop true */
          }
        }  

        /* Gateway IP address */
        if (LINE_COUNTER == 6)                    /* Gateway IP address ? */
        {
          #ifdef _DEBUG_
            Serial.print ("    Gateway IP-Address: ");
          # endif 
          TEST_INT = IP_ADRESS_STRING.toInt();
          /* In range ? */
          if ((TEST_INT >= 0) && (TEST_INT <= 255))
          {
            GATEWAY[3] = TEST_INT;                /* Save 3rd digit */
            IP_ADDRESS_COUNTER = 5;               /* Set to fifth address */                            
          }
          else                                    /* Not in range ? */
          {
            LOOP_COUNTER = 50;                    /* Set loop counter out of range */
          }
          /* Check if 4 digits received and loop counter in range */
          if ((LOOP_COUNTER < 20) && (IP_ADDRESS_COUNTER == 5))        
          {
            #ifdef _DEBUG_
              Serial.println (GATEWAY);
            # endif 
            LINE_COUNTER = 7;                     /* Set line counter to seventh line */
            IP_ADDRESS_COUNTER = 1;               /* Set to first address */
            LOOP_COUNTER = 0;                     /* Reset loop counter */
            IP_ADRESS_STRING = "";                /* Delete string for next digit */
          }
          else                                    /* Error ? */  
          {
            #ifdef _DEBUG_
              Serial.println ("Gateway IP-Address format error!");
            # endif 
            BREAK_LOOP = true;                    /* Set break loop true */
          }
        } 

       /* Subnet mask */
        if (LINE_COUNTER == 5)                    /* Subnet mask ? */
        {
          #ifdef _DEBUG_
            Serial.print ("    Subnet mask: ");
          # endif 
          TEST_INT = IP_ADRESS_STRING.toInt();
          /* In range ? */
          if ((TEST_INT >= 0) && (TEST_INT <= 255))
          {
            SUBNET[3] = TEST_INT;                 /* Save 3rd digit */   
            IP_ADDRESS_COUNTER = 5;               /* Set to fifth address */            
          }
          else                                    /* Not in range ? */
          {
            LOOP_COUNTER = 50;                    /* Set loop counter out of range */
          }
          /* Check if 4 digits received and loop counter in range */
          if ((LOOP_COUNTER < 20) && (IP_ADDRESS_COUNTER == 5))        
          {
            #ifdef _DEBUG_
              Serial.println (SUBNET);
            # endif 
            LINE_COUNTER = 6;                     /* Set line counter to sixth line */
            IP_ADDRESS_COUNTER = 1;               /* Set to first address */
            LOOP_COUNTER = 0;                     /* Reset loop counter */
            IP_ADRESS_STRING = "";                /* Delete string for next digit */
          }
          else                                    /* Error ? */  
          {
            #ifdef _DEBUG_
              Serial.println ("Subnet mask format error!");
            # endif 
            BREAK_LOOP = true;                    /* Set break loop true */
          }
        }              

        
        /* Fixed IP address */
        if (LINE_COUNTER == 4)                    /* Fixed IP address ? */
        {
          #ifdef _DEBUG_
            Serial.print ("    Fixed IP-Address: ");
          # endif 
          TEST_INT = IP_ADRESS_STRING.toInt();
          /* In range ? */
          if ((TEST_INT >= 0) && (TEST_INT <= 255))
          {
            FIXED_IP[3] = TEST_INT;               /* Save 3rd digit */ 
            IP_ADDRESS_COUNTER = 5;               /* Set to fifth address */             
          }
          else                                    /* Not in range ? */
          {
            LOOP_COUNTER = 50;                    /* Set loop counter out of range */
          }
          /* Check if 4 digits received and loop counter in range */
          if ((LOOP_COUNTER < 20) && (IP_ADDRESS_COUNTER == 5))        
          {
            #ifdef _DEBUG_
              Serial.println (FIXED_IP);
            # endif 
            LINE_COUNTER = 5;                     /* Set line counter to fifth line */
            IP_ADDRESS_COUNTER = 1;               /* Set to first address */
            LOOP_COUNTER = 0;                     /* Reset loop counter */
            IP_ADRESS_STRING = "";                /* Delete string for next digit */
          }
          else                                    /* Error ? */  
          {
            #ifdef _DEBUG_
              Serial.println ("Fixed IP address format error!");
            # endif 
            BREAK_LOOP = true;                    /* Set break loop true */
          }
        }
        
        /* Hostname */
        if (LINE_COUNTER == 3)                    /* Receive Hostname ? */
        {
          #ifdef _DEBUG_
            Serial.print ("    Hostname: ");
          # endif 
          /* Check on too long hostname */
          if ((LOOP_COUNTER > 1) && (LOOP_COUNTER < 20))        
          {
            #ifdef _DEBUG_
              Serial.println (HOSTNAME);
            # endif 
            LINE_COUNTER = 4;                     /* Set line counter to forth line */
            LOOP_COUNTER = 0;                     /* Reset loop counter */
            IP_ADDRESS_COUNTER = 1;               /* Set to first address */
            IP_ADRESS_STRING = "";                /* Delete string */            
          }
          else                                    /* SSID length error ? */
          {
             #ifdef _DEBUG_
              Serial.println ("Hostname length error!");
            # endif 
            BREAK_LOOP = true;                    /* Set break loop true */
          }
        }
        
        /* WIFI Password */
        if (LINE_COUNTER == 2)                    /* Receive WIFI password ? */
        {
          #ifdef _DEBUG_
            Serial.print ("    WIFI password: ");
          # endif 
          /* Check on too long WIFI password */
          if ((LOOP_COUNTER > 1) && (LOOP_COUNTER < 20))        
          {
            #ifdef _DEBUG_
              Serial.println (WLAN_PWD);
            # endif 
            LINE_COUNTER = 3;                     /* Set line counter to third line */
            LOOP_COUNTER = 0;                     /* Reset loop counter */            
          }
          else                                    /* WIFI password length error ? */
          {
             #ifdef _DEBUG_
              Serial.println ("WIFI password length error!");
            # endif 
            BREAK_LOOP = true;                    /* Set break loop true */
          }
        }
        
        /* SSID */
        if (LINE_COUNTER == 1)                    /* Receive SSID ? */
        {
          #ifdef _DEBUG_
            Serial.print ("    WIFI SSID: ");
          # endif 
          /* Check on too long SSID */
          if ((LOOP_COUNTER > 1) && (LOOP_COUNTER < 20))        
          {
            #ifdef _DEBUG_
              Serial.println (WLAN_SSID);
            # endif 
            LINE_COUNTER = 2;                     /* Set line counter to second line */
            LOOP_COUNTER = 0;                     /* Reset loop counter */            
          }
          else                                    /* SSID length error ? */
          {
             #ifdef _DEBUG_
              Serial.println ("SSID length error!");
            # endif 
            BREAK_LOOP = true;                    /* Set break loop true */
          }
        }   
      }
      else                                        /* No CR received ? */
      {
        if (RECEIVE_CHAR != '\n')                 /* No line feed ? */
        {
          if (LOOP_COUNTER < 20) 
          {
            if (LINE_COUNTER == 1)                /* Receive SSID ? */
            {
              /* Add received char */
              WLAN_SSID[LOOP_COUNTER] = RECEIVE_CHAR;      
            }
            if (LINE_COUNTER == 2)                /* Receive WIFI password ? */
            {
              /* Add received char */
              WLAN_PWD[LOOP_COUNTER] = RECEIVE_CHAR;      
            }
            if (LINE_COUNTER == 3)                /* Receive Hostname ? */
            {
              /* Add received char */
              HOSTNAME[LOOP_COUNTER] = RECEIVE_CHAR;      
            }
            
            if (LINE_COUNTER == 4)                /* Receive fixed IP address ? */
            { 
              if (RECEIVE_CHAR != '.')            /* Delimiter not detected ? */
              {
                IP_ADRESS_STRING += RECEIVE_CHAR; /* Add char */
              }
              else                                /* Delimiter detected ? */
              {             
                if (IP_ADDRESS_COUNTER == 3)      /* 3rd digit ? */
                {
                  /* Convert to Int */
                  TEST_INT = IP_ADRESS_STRING.toInt();
                  /* In range ? */
                  if ((TEST_INT >= 0) && (TEST_INT <= 255))
                  {
                    FIXED_IP[2] = TEST_INT;       /* Save 3rd digit */               
                    IP_ADDRESS_COUNTER = 4;       /* Set to forth digit */
                    IP_ADRESS_STRING = "";        /* Delete string for next digit */
                  }
                  else                            /* Not in range ? */
                  {
                    LOOP_COUNTER = 50;            /* Set loop counter out of range */
                  }
                }
                if (IP_ADDRESS_COUNTER == 2)      /* 2nd digit ? */
                {
                  /* Convert to Int */
                  TEST_INT = IP_ADRESS_STRING.toInt();
                  /* In range ? */
                  if ((TEST_INT >= 0) && (TEST_INT <= 255))
                  {
                    FIXED_IP[1] = TEST_INT;       /* Save 2rd digit */               
                    IP_ADDRESS_COUNTER = 3;       /* Set to third digit */
                    IP_ADRESS_STRING = "";        /* Delete string for next digit */
                  }
                  else                            /* Not in range ? */
                  {
                    LOOP_COUNTER = 50;            /* Set loop counter out of range */
                  }
                }                
                if (IP_ADDRESS_COUNTER == 1)      /* 1st digit ? */
                {
                  /* Convert to Int */
                  TEST_INT = IP_ADRESS_STRING.toInt();
                  /* In range ? */
                  if ((TEST_INT >= 0) && (TEST_INT <= 255))
                  {
                    FIXED_IP[0] = TEST_INT;       /* Save 1st digit */               
                    IP_ADDRESS_COUNTER = 2;       /* Set to second digit */
                    IP_ADRESS_STRING = "";        /* Delete string for next digit */
                  }
                  else                            /* Not in range ? */
                  {
                    LOOP_COUNTER = 50;            /* Set loop counter out of range */
                  }
                }  
              }    
            }

            if (LINE_COUNTER == 5)                /* Subnet mask ? */
            { 
              if (RECEIVE_CHAR != '.')            /* Delimiter not detected ? */
              {
                IP_ADRESS_STRING += RECEIVE_CHAR; /* Add char */
              }
              else                                /* Delimiter detected ? */
              {             
                if (IP_ADDRESS_COUNTER == 3)      /* 3rd digit ? */
                {
                  /* Convert to Int */
                  TEST_INT = IP_ADRESS_STRING.toInt();
                  /* In range ? */
                  if ((TEST_INT >= 0) && (TEST_INT <= 255))
                  {
                    SUBNET[2] = TEST_INT;         /* Save 3rd digit */               
                    IP_ADDRESS_COUNTER = 4;       /* Set to forth digit */
                    IP_ADRESS_STRING = "";        /* Delete string for next digit */
                  }
                  else                            /* Not in range ? */
                  {
                    LOOP_COUNTER = 50;            /* Set loop counter out of range */
                  }
                }
                if (IP_ADDRESS_COUNTER == 2)      /* 2nd digit ? */
                {
                  /* Convert to Int */
                  TEST_INT = IP_ADRESS_STRING.toInt();
                  /* In range ? */
                  if ((TEST_INT >= 0) && (TEST_INT <= 255))
                  {
                    SUBNET[1] = TEST_INT;         /* Save 2rd digit */               
                    IP_ADDRESS_COUNTER = 3;       /* Set to third digit */
                    IP_ADRESS_STRING = "";        /* Delete string for next digit */
                  }
                  else                            /* Not in range ? */
                  {
                    LOOP_COUNTER = 50;            /* Set loop counter out of range */
                  }
                }                
                if (IP_ADDRESS_COUNTER == 1)      /* 1st digit ? */
                {
                  /* Convert to Int */
                  TEST_INT = IP_ADRESS_STRING.toInt();
                  /* In range ? */
                  if ((TEST_INT >= 0) && (TEST_INT <= 255))
                  {
                    SUBNET[0] = TEST_INT;         /* Save 1st digit */               
                    IP_ADDRESS_COUNTER = 2;       /* Set to second digit */
                    IP_ADRESS_STRING = "";        /* Delete string for next digit */
                  }
                  else                            /* Not in range ? */
                  {
                    LOOP_COUNTER = 50;            /* Set loop counter out of range */
                  }
                }  
              }    
            }            

            if (LINE_COUNTER == 6)                /* Gateway address ? */
            { 
              if (RECEIVE_CHAR != '.')            /* Delimiter not detected ? */
              {
                IP_ADRESS_STRING += RECEIVE_CHAR; /* Add char */
              }
              else                                /* Delimiter detected ? */
              {             
                if (IP_ADDRESS_COUNTER == 3)      /* 3rd digit ? */
                {
                  /* Convert to Int */
                  TEST_INT = IP_ADRESS_STRING.toInt();
                  /* In range ? */
                  if ((TEST_INT >= 0) && (TEST_INT <= 255))
                  {
                    GATEWAY[2] = TEST_INT;        /* Save 3rd digit */               
                    IP_ADDRESS_COUNTER = 4;       /* Set to forth digit */
                    IP_ADRESS_STRING = "";        /* Delete string for next digit */
                  }
                  else                            /* Not in range ? */
                  {
                    LOOP_COUNTER = 50;            /* Set loop counter out of range */
                  }
                }
                if (IP_ADDRESS_COUNTER == 2)      /* 2nd digit ? */
                {
                  /* Convert to Int */
                  TEST_INT = IP_ADRESS_STRING.toInt();
                  /* In range ? */
                  if ((TEST_INT >= 0) && (TEST_INT <= 255))
                  {
                    GATEWAY[1] = TEST_INT;        /* Save 2rd digit */               
                    IP_ADDRESS_COUNTER = 3;       /* Set to third digit */
                    IP_ADRESS_STRING = "";        /* Delete string for next digit */
                  }
                  else                            /* Not in range ? */
                  {
                    LOOP_COUNTER = 50;            /* Set loop counter out of range */
                  }
                }                
                if (IP_ADDRESS_COUNTER == 1)      /* 1st digit ? */
                {
                  /* Convert to Int */
                  TEST_INT = IP_ADRESS_STRING.toInt();
                  /* In range ? */
                  if ((TEST_INT >= 0) && (TEST_INT <= 255))
                  {
                    GATEWAY[0] = TEST_INT;        /* Save 1st digit */               
                    IP_ADDRESS_COUNTER = 2;       /* Set to second digit */
                    IP_ADRESS_STRING = "";        /* Delete string for next digit */
                  }
                  else                            /* Not in range ? */
                  {
                    LOOP_COUNTER = 50;            /* Set loop counter out of range */
                  }
                }  
              }    
            }


            if (LINE_COUNTER == 7)                /* DNS 1 address ? */
            { 
              if (RECEIVE_CHAR != '.')            /* Delimiter not detected ? */
              {
                IP_ADRESS_STRING += RECEIVE_CHAR; /* Add char */
              }
              else                                /* Delimiter detected ? */
              {             
                if (IP_ADDRESS_COUNTER == 3)      /* 3rd digit ? */
                {
                  /* Convert to Int */
                  TEST_INT = IP_ADRESS_STRING.toInt();
                  /* In range ? */
                  if ((TEST_INT >= 0) && (TEST_INT <= 255))
                  {
                    DNS_1[2] = TEST_INT;          /* Save 3rd digit */               
                    IP_ADDRESS_COUNTER = 4;       /* Set to forth digit */
                    IP_ADRESS_STRING = "";        /* Delete string for next digit */
                  }
                  else                            /* Not in range ? */
                  {
                    LOOP_COUNTER = 50;            /* Set loop counter out of range */
                  }
                }
                if (IP_ADDRESS_COUNTER == 2)      /* 2nd digit ? */
                {
                  /* Convert to Int */
                  TEST_INT = IP_ADRESS_STRING.toInt();
                  /* In range ? */
                  if ((TEST_INT >= 0) && (TEST_INT <= 255))
                  {
                    DNS_1[1] = TEST_INT;          /* Save 2rd digit */               
                    IP_ADDRESS_COUNTER = 3;       /* Set to third digit */
                    IP_ADRESS_STRING = "";        /* Delete string for next digit */
                  }
                  else                            /* Not in range ? */
                  {
                    LOOP_COUNTER = 50;            /* Set loop counter out of range */
                  }
                }                
                if (IP_ADDRESS_COUNTER == 1)      /* 1st digit ? */
                {
                  /* Convert to Int */
                  TEST_INT = IP_ADRESS_STRING.toInt();
                  /* In range ? */
                  if ((TEST_INT >= 0) && (TEST_INT <= 255))
                  {
                    DNS_1[0] = TEST_INT;          /* Save 1st digit */               
                    IP_ADDRESS_COUNTER = 2;       /* Set to second digit */
                    IP_ADRESS_STRING = "";        /* Delete string for next digit */
                  }
                  else                            /* Not in range ? */
                  {
                    LOOP_COUNTER = 50;            /* Set loop counter out of range */
                  }
                }  
              }    
            }

            if (LINE_COUNTER == 8)                /* DNS 2 address ? */
            { 
              if (RECEIVE_CHAR != '.')            /* Delimiter not detected ? */
              {
                IP_ADRESS_STRING += RECEIVE_CHAR; /* Add char */
              }
              else                                /* Delimiter detected ? */
              {             
                if (IP_ADDRESS_COUNTER == 3)      /* 3rd digit ? */
                {
                  /* Convert to Int */
                  TEST_INT = IP_ADRESS_STRING.toInt();
                  /* In range ? */
                  if ((TEST_INT >= 0) && (TEST_INT <= 255))
                  {
                    DNS_2[2] = TEST_INT;          /* Save 3rd digit */               
                    IP_ADDRESS_COUNTER = 4;       /* Set to forth digit */
                    IP_ADRESS_STRING = "";        /* Delete string for next digit */
                  }
                  else                            /* Not in range ? */
                  {
                    LOOP_COUNTER = 50;            /* Set loop counter out of range */
                  }
                }
                if (IP_ADDRESS_COUNTER == 2)      /* 2nd digit ? */
                {
                  /* Convert to Int */
                  TEST_INT = IP_ADRESS_STRING.toInt();
                  /* In range ? */
                  if ((TEST_INT >= 0) && (TEST_INT <= 255))
                  {
                    DNS_2[1] = TEST_INT;          /* Save 2rd digit */               
                    IP_ADDRESS_COUNTER = 3;       /* Set to third digit */
                    IP_ADRESS_STRING = "";        /* Delete string for next digit */
                  }
                  else                            /* Not in range ? */
                  {
                    LOOP_COUNTER = 50;            /* Set loop counter out of range */
                  }
                }                
                if (IP_ADDRESS_COUNTER == 1)      /* 1st digit ? */
                {
                  /* Convert to Int */
                  TEST_INT = IP_ADRESS_STRING.toInt();
                  /* In range ? */
                  if ((TEST_INT >= 0) && (TEST_INT <= 255))
                  {
                    DNS_2[0] = TEST_INT;          /* Save 1st digit */               
                    IP_ADDRESS_COUNTER = 2;       /* Set to second digit */
                    IP_ADRESS_STRING = "";        /* Delete string for next digit */
                  }
                  else                            /* Not in range ? */
                  {
                    LOOP_COUNTER = 50;            /* Set loop counter out of range */
                  }
                }  
              }    
            }

            if (LINE_COUNTER == 9)                /* Receive NTP server 1 ? */
            {
              /* Add received char */
              NTP_SERVER_1[LOOP_COUNTER] = RECEIVE_CHAR;      
            }
            
            if (LINE_COUNTER == 10)               /* Receive NTP server 2 ? */
            {
              /* Add received char */
              NTP_SERVER_2[LOOP_COUNTER] = RECEIVE_CHAR;      
            }        
                
            if (LINE_COUNTER == 11)               /* Receive NTP server 3 ? */
            {
              /* Add received char */
              NTP_SERVER_3[LOOP_COUNTER] = RECEIVE_CHAR;      
            } 
                        
            LOOP_COUNTER++;                       /* Increment loop counter */
          }
          else                                    /* Length error */
          {
            #ifdef _DEBUG_
              Serial.println ("File length error!!!");
            # endif 
            BREAK_LOOP = true;                    /* Set break loop true */          
          }
        }
      }
    }
    file.close();                                 /* Close file after reading */
    
    if (BREAK_LOOP == false)
    {   
      /* NTP 3 server */
      #ifdef _DEBUG_
        Serial.print ("    NTP 3 server: ");
      # endif 
      /* Check on too long NTP 1 server name */
      if ((LOOP_COUNTER > 1) && (LOOP_COUNTER < 20))        
      {
        #ifdef _DEBUG_
          Serial.println (NTP_SERVER_3);
        # endif            
      }
      else                                        /* NTP 3 length error */
      {
        #ifdef _DEBUG_
          Serial.println ("NTP server 2 length error!");
        # endif 
        BREAK_LOOP = true;                        /* Set break loop true */
      }
    }     
  }


  if (BREAK_LOOP == true)                         /* Error ? */
  {
    do                                            /* Endless loop until manual reset */
    {
      SUB_TRIGGER_WATCHDOG ();                    /* Trigger watchdog */
      delay (500);                                /* 500ms delay */
    }
    while (1);
  }
}

/****************************************************************************************/
/* Subroutine to save settings in init file Init.txt                                    */
/****************************************************************************************/
void SUB_WRITE_INIT_FILE (void)
{ 
  Serial.print ("  Remove old Init.txt: ");
  if(SPIFFS.remove("/Init.txt"))                  /* Remove existing init file successful ? */
  {
    Serial.println("Done");                       /* Print success mesage */                
  }
  else                                            /* Remove existing init file not successful ? */
  {
    Serial.println("Failed");                     /* Print failed mesage */ 
  }

  Serial.print ("  Open Init.txt: ");
  /* Open init file for write access */
  File FileToWrite = SPIFFS.open("/Init.txt", FILE_WRITE);
  if(!FileToWrite)                                /* File opening was not successful ? */
  {
    Serial.println("Failed!");                    /* Print failed mesage */ 
    return;                                       /* Stop executing sub routine */
  }
  else                                            /* File opening was successful ? */
  {
    Serial.println("Done");
    Serial.print("  Writing Init.txt file: ");  
    FileToWrite.println(WLAN_SSID);               /* Write WIFI SSID */
    FileToWrite.println(WLAN_PWD);                /* Write WIFI password */
    FileToWrite.println(HOSTNAME);                /* Write hostname */
    FileToWrite.println(FIXED_IP);                /* Write Fixed IP-Address */
    FileToWrite.println(SUBNET);                  /* Write Subnet mask */
    FileToWrite.println(GATEWAY);                 /* Write Gateway IP-Address */
    FileToWrite.println(DNS_1);                   /* Write DNS 1 IP-Address */
    FileToWrite.println(DNS_2);                   /* Write DNS 2 IP-Address */    
    FileToWrite.println(NTP_SERVER_1);            /* Write NTP 1 server name */
    FileToWrite.println(NTP_SERVER_2);            /* Write NTP 2 server name */
    FileToWrite.println(NTP_SERVER_3);            /* Write NTP 3 server name */                      
    FileToWrite.close();                          /* Close file */  
    Serial.println("Done");
  }
}


/****************************************************************************************/
/* Subroutine to change configuration settings via console entry                        */
/****************************************************************************************/
void SUB_CONFIG_SETTING (void)
{
  char RECEIVE_CHAR;                              /* Received byte from file reading */
  byte RECEIVE_BYTE_COUNT;                        /* Received byte counter */ 
  char INPUT_STRING[21] = {0};                    /* Input string */
  char YES_NO_BUFFER[2];                          /* Buffer for Yes/No decision */
  
  /* Display welcome configuration message with basic information */
  Serial.println("\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n");
  Serial.println ("*********************************************************");
  Serial.println ("* Configuration of initial connection parameters        *");
  Serial.println ("* Press enter if parameter should remain as is          *");
  Serial.println ("* Please enter IP addresses in format: 192,168,123,219  *");   
  Serial.println ("*********************************************************\n"); 

  delay (500);                                    /* 500 ms delay */
 
  while( Serial.available())                      /* Clear incoming serial buffer */
  { 
    RECEIVE_CHAR = Serial.read();                 /* Read char if available */
  }  

  /* Step 1: Change actual WIFI SSID */
  Serial.print ("Actual WIFI SSID is \"");
  Serial.print (WLAN_SSID);
  Serial.print ("\", enter new value (<20 chars) or press <ENTER> to keep current name: ");
  memset(INPUT_STRING, 0, sizeof(INPUT_STRING));  /* Clear contents of input buffer */
  RECEIVE_BYTE_COUNT = 0;                         /* Reset byte counter */
  RECEIVE_CHAR = 32;                              /* Delete received char */
  do
  {
    if (Serial.available()>0)                     /* New character available ? */
    {
      RECEIVE_CHAR = Serial.read();               /* Read char if available */
      /* No CR and no LF received */
      if ((RECEIVE_CHAR != '\n') && (RECEIVE_CHAR != '\r'))
      {
        /* Save char */
        INPUT_STRING [RECEIVE_BYTE_COUNT] = RECEIVE_CHAR;
        Serial.print (RECEIVE_CHAR);
        RECEIVE_BYTE_COUNT++;                     /* Increase byte counter */
      }
    }
  }
  while ((RECEIVE_CHAR != '\n') && (RECEIVE_CHAR != '\r') && (RECEIVE_BYTE_COUNT < 20));

  /* Not ENTER pressed and string not too long --> Save value */
  if ((RECEIVE_BYTE_COUNT > 0) && (RECEIVE_BYTE_COUNT < 20))                         
  {
    memset(WLAN_SSID, 0, sizeof(WLAN_SSID));      /* Clear contents of WLAN_SSID */ 
    strcat(WLAN_SSID, INPUT_STRING);              /* Copy string */
  }
  Serial.println ();
  Serial.print ("  WIFI SSID set to \"");
  Serial.print (WLAN_SSID);
  Serial.println ("\"");

  /* Step 2: Change actual WIFI Password */
  delay (500);                                    /* 500 ms delay */
  while( Serial.available())                      /* Clear incoming serial buffer */
  { 
    RECEIVE_CHAR = Serial.read();                 /* Read char if available */
  }  
  Serial.print ("Actual WIFI password is \"");
  Serial.print (WLAN_PWD);
  Serial.print ("\", enter new value (<20 chars) or press <ENTER> to keep current name: ");
  memset(INPUT_STRING, 0, sizeof(INPUT_STRING));  /* Clear contents of input buffer */
  RECEIVE_BYTE_COUNT = 0;                         /* Reset byte counter */
  RECEIVE_CHAR = 32;                              /* Delete received char */
  do
  {
    if (Serial.available()>0)                     /* New character available ? */
    {
      RECEIVE_CHAR = Serial.read();               /* Read char if available */
      /* No CR and no LF received */
      if ((RECEIVE_CHAR != '\n') && (RECEIVE_CHAR != '\r'))
      {
        /* Save char */
        INPUT_STRING [RECEIVE_BYTE_COUNT] = RECEIVE_CHAR;
        Serial.print (RECEIVE_CHAR);
        RECEIVE_BYTE_COUNT++;                     /* Increase byte counter */
      }
    }
  }
  while ((RECEIVE_CHAR != '\n') && (RECEIVE_CHAR != '\r') && (RECEIVE_BYTE_COUNT < 20));

  /* Not ENTER pressed and string not too long --> Save value */
  if ((RECEIVE_BYTE_COUNT > 0) && (RECEIVE_BYTE_COUNT < 20))                         
  {
    memset(WLAN_PWD, 0, sizeof(WLAN_PWD));        /* Clear contents of WLAN_SSID */ 
    strcat(WLAN_PWD, INPUT_STRING);               /* Copy string */
  }
  Serial.println ();
  Serial.print ("  WIFI password set to \"");
  Serial.print (WLAN_PWD);
  Serial.println ("\"");  

  /* Step 3: Change actual Hostname */
  delay (500);                                    /* 500 ms delay */
  while( Serial.available())                      /* Clear incoming serial buffer */
  { 
    RECEIVE_CHAR = Serial.read();                 /* Read char if available */
  }  
  Serial.print ("Actual Hostname is \"");
  Serial.print (HOSTNAME);
  Serial.print ("\", enter new value (<20 chars) or press <ENTER> to keep current name: ");
  memset(INPUT_STRING, 0, sizeof(INPUT_STRING));  /* Clear contents of input buffer */
  RECEIVE_BYTE_COUNT = 0;                         /* Reset byte counter */
  RECEIVE_CHAR = 32;                              /* Delete received char */
  do
  {
    if (Serial.available()>0)                     /* New character available ? */
    {
      RECEIVE_CHAR = Serial.read();               /* Read char if available */
      /* No CR and no LF received */
      if ((RECEIVE_CHAR != '\n') && (RECEIVE_CHAR != '\r'))
      {
        /* Save char */
        INPUT_STRING [RECEIVE_BYTE_COUNT] = RECEIVE_CHAR;
        Serial.print (RECEIVE_CHAR);
        RECEIVE_BYTE_COUNT++;                     /* Increase byte counter */
      }
    }
  }
  while ((RECEIVE_CHAR != '\n') && (RECEIVE_CHAR != '\r') && (RECEIVE_BYTE_COUNT < 20));

  /* Not ENTER pressed and string not too long --> Save value */
  if ((RECEIVE_BYTE_COUNT > 0) && (RECEIVE_BYTE_COUNT < 20))                         
  {
    memset(HOSTNAME, 0, sizeof(HOSTNAME));        /* Clear contents of WLAN_SSID */ 
    strcat(HOSTNAME, INPUT_STRING);               /* Copy string */
  }
  Serial.println ();
  Serial.print ("  Hostname set to \"");
  Serial.print (HOSTNAME);
  Serial.println ("\"");    

  /* Step 4: Change actual IP-Adress */
  delay (500);                                    /* 500 ms delay */
  while( Serial.available())                      /* Clear incoming serial buffer */
  { 
    RECEIVE_CHAR = Serial.read();                 /* Read char if available */
  }  
  Serial.print ("Actual Fixed IP-Address is \"");
  Serial.print (FIXED_IP);
  Serial.print ("\", enter new value (<20 chars) or press <ENTER> to keep current name: ");
  memset(INPUT_STRING, 0, sizeof(INPUT_STRING));  /* Clear contents of input buffer */
  RECEIVE_BYTE_COUNT = 0;                         /* Reset byte counter */
  RECEIVE_CHAR = 32;                              /* Delete received char */
  do
  {
    if (Serial.available()>0)                     /* New character available ? */
    {
      RECEIVE_CHAR = Serial.read();               /* Read char if available */
      /* No CR and no LF received */
      if ((RECEIVE_CHAR != '\n') && (RECEIVE_CHAR != '\r'))
      {
        /* Save char */
        INPUT_STRING [RECEIVE_BYTE_COUNT] = RECEIVE_CHAR;
        Serial.print (RECEIVE_CHAR);
        RECEIVE_BYTE_COUNT++;                     /* Increase byte counter */
      }
    }
  }
  while ((RECEIVE_CHAR != '\n') && (RECEIVE_CHAR != '\r') && (RECEIVE_BYTE_COUNT < 20));

  /* Not ENTER pressed and string not too long --> Save value */
  if ((RECEIVE_BYTE_COUNT > 0) && (RECEIVE_BYTE_COUNT < 20))                         
  {
    FIXED_IP.fromString(INPUT_STRING);
  } 

  Serial.println ();
  Serial.print ("  Fixed IP-Adress set to \"");
  Serial.print (FIXED_IP);
  Serial.println ("\"");    

  /* Step 5: Change actual Subnet mask */
  delay (500);                                    /* 500 ms delay */
  while( Serial.available())                      /* Clear incoming serial buffer */
  { 
    RECEIVE_CHAR = Serial.read();                 /* Read char if available */
  }  
  Serial.print ("Actual Subnet mask is \"");
  Serial.print (SUBNET);
  Serial.print ("\", enter new value (<20 chars) or press <ENTER> to keep current name: ");
  memset(INPUT_STRING, 0, sizeof(INPUT_STRING));  /* Clear contents of input buffer */
  RECEIVE_BYTE_COUNT = 0;                         /* Reset byte counter */
  RECEIVE_CHAR = 32;                              /* Delete received char */
  do
  {
    if (Serial.available()>0)                     /* New character available ? */
    {
      RECEIVE_CHAR = Serial.read();               /* Read char if available */
      /* No CR and no LF received */
      if ((RECEIVE_CHAR != '\n') && (RECEIVE_CHAR != '\r'))
      {
        /* Save char */
        INPUT_STRING [RECEIVE_BYTE_COUNT] = RECEIVE_CHAR;
        Serial.print (RECEIVE_CHAR);
        RECEIVE_BYTE_COUNT++;                     /* Increase byte counter */
      }
    }
  }
  while ((RECEIVE_CHAR != '\n') && (RECEIVE_CHAR != '\r') && (RECEIVE_BYTE_COUNT < 20));

  /* Not ENTER pressed and string not too long --> Save value */
  if ((RECEIVE_BYTE_COUNT > 0) && (RECEIVE_BYTE_COUNT < 20))                         
  {
    SUBNET.fromString(INPUT_STRING);
  } 

  Serial.println ();
  Serial.print ("  Subnet mask set to \"");
  Serial.print (SUBNET);
  Serial.println ("\"");  

  /* Step 6: Change actual Gateway IP-Address */
  delay (500);                                    /* 500 ms delay */
  while( Serial.available())                      /* Clear incoming serial buffer */
  { 
    RECEIVE_CHAR = Serial.read();                 /* Read char if available */
  }  
  Serial.print ("Actual Gateway IP-Address is \"");
  Serial.print (GATEWAY);
  Serial.print ("\", enter new value (<20 chars) or press <ENTER> to keep current name: ");
  memset(INPUT_STRING, 0, sizeof(INPUT_STRING));  /* Clear contents of input buffer */
  RECEIVE_BYTE_COUNT = 0;                         /* Reset byte counter */
  RECEIVE_CHAR = 32;                              /* Delete received char */
  do
  {
    if (Serial.available()>0)                     /* New character available ? */
    {
      RECEIVE_CHAR = Serial.read();               /* Read char if available */
      /* No CR and no LF received */
      if ((RECEIVE_CHAR != '\n') && (RECEIVE_CHAR != '\r'))
      {
        /* Save char */
        INPUT_STRING [RECEIVE_BYTE_COUNT] = RECEIVE_CHAR;
        Serial.print (RECEIVE_CHAR);
        RECEIVE_BYTE_COUNT++;                     /* Increase byte counter */
      }
    }
  }
  while ((RECEIVE_CHAR != '\n') && (RECEIVE_CHAR != '\r') && (RECEIVE_BYTE_COUNT < 20));

  /* Not ENTER pressed and string not too long --> Save value */
  if ((RECEIVE_BYTE_COUNT > 0) && (RECEIVE_BYTE_COUNT < 20))                         
  {
    GATEWAY.fromString(INPUT_STRING);
  } 

  Serial.println ();
  Serial.print ("  Gateway IP-Adress set to \"");
  Serial.print (GATEWAY);
  Serial.println ("\"");  

  /* Step 7: Change actual DNS server 1 address */
  delay (500);                                    /* 500 ms delay */
  while( Serial.available())                      /* Clear incoming serial buffer */
  { 
    RECEIVE_CHAR = Serial.read();                 /* Read char if available */
  }  
  Serial.print ("Actual DNS 1 IP-Address is \"");
  Serial.print (DNS_1);
  Serial.print ("\", enter new value (<20 chars) or press <ENTER> to keep current name: ");
  memset(INPUT_STRING, 0, sizeof(INPUT_STRING));  /* Clear contents of input buffer */
  RECEIVE_BYTE_COUNT = 0;                         /* Reset byte counter */
  RECEIVE_CHAR = 32;                              /* Delete received char */
  do
  {
    if (Serial.available()>0)                     /* New character available ? */
    {
      RECEIVE_CHAR = Serial.read();               /* Read char if available */
      /* No CR and no LF received */
      if ((RECEIVE_CHAR != '\n') && (RECEIVE_CHAR != '\r'))
      {
        /* Save char */
        INPUT_STRING [RECEIVE_BYTE_COUNT] = RECEIVE_CHAR;
        Serial.print (RECEIVE_CHAR);
        RECEIVE_BYTE_COUNT++;                     /* Increase byte counter */
      }
    }
  }
  while ((RECEIVE_CHAR != '\n') && (RECEIVE_CHAR != '\r') && (RECEIVE_BYTE_COUNT < 20));

  /* Not ENTER pressed and string not too long --> Save value */
  if ((RECEIVE_BYTE_COUNT > 0) && (RECEIVE_BYTE_COUNT < 20))                         
  {
    DNS_1.fromString(INPUT_STRING);
  } 
  Serial.println ();
  Serial.print ("  DNS 1 IP-Adress set to \"");
  Serial.print (DNS_1);
  Serial.println ("\""); 

  /* Step 8: Change actual DNS server 2 address */
  delay (500);                                    /* 500 ms delay */
  while( Serial.available())                      /* Clear incoming serial buffer */
  { 
    RECEIVE_CHAR = Serial.read();                 /* Read char if available */
  }  
  Serial.print ("Actual DNS 2 IP-Address is \"");
  Serial.print (DNS_2);
  Serial.print ("\", enter new value (<20 chars) or press <ENTER> to keep current name: ");
  memset(INPUT_STRING, 0, sizeof(INPUT_STRING));  /* Clear contents of input buffer */
  RECEIVE_BYTE_COUNT = 0;                         /* Reset byte counter */
  RECEIVE_CHAR = 32;                              /* Delete received char */
  do
  {
    if (Serial.available()>0)                     /* New character available ? */
    {
      RECEIVE_CHAR = Serial.read();               /* Read char if available */
      /* No CR and no LF received */
      if ((RECEIVE_CHAR != '\n') && (RECEIVE_CHAR != '\r'))
      {
        /* Save char */
        INPUT_STRING [RECEIVE_BYTE_COUNT] = RECEIVE_CHAR;
        Serial.print (RECEIVE_CHAR);
        RECEIVE_BYTE_COUNT++;                     /* Increase byte counter */
      }
    }
  }
  while ((RECEIVE_CHAR != '\n') && (RECEIVE_CHAR != '\r') && (RECEIVE_BYTE_COUNT < 20));

  /* Not ENTER pressed and string not too long --> Save value */
  if ((RECEIVE_BYTE_COUNT > 0) && (RECEIVE_BYTE_COUNT < 20))                         
  {
    DNS_2.fromString(INPUT_STRING);
  } 
  Serial.println ();
  Serial.print ("  DNS 2 IP-Adress set to \"");
  Serial.print (DNS_2);
  Serial.println ("\"");

  /* Step 9: Change actual NTP 1 server */
  delay (500);                                    /* 500 ms delay */
  while( Serial.available())                      /* Clear incoming serial buffer */
  { 
    RECEIVE_CHAR = Serial.read();                 /* Read char if available */
  }  
  Serial.print ("Actual NTP 1 name is \"");
  Serial.print (NTP_SERVER_1);
  Serial.print ("\", enter new value (<20 chars) or press <ENTER> to keep current name: ");
  memset(INPUT_STRING, 0, sizeof(INPUT_STRING));  /* Clear contents of input buffer */
  RECEIVE_BYTE_COUNT = 0;                         /* Reset byte counter */
  RECEIVE_CHAR = 32;                              /* Delete received char */
  do
  {
    if (Serial.available()>0)                     /* New character available ? */
    {
      RECEIVE_CHAR = Serial.read();               /* Read char if available */
      /* No CR and no LF received */
      if ((RECEIVE_CHAR != '\n') && (RECEIVE_CHAR != '\r'))
      {
        /* Save char */
        INPUT_STRING [RECEIVE_BYTE_COUNT] = RECEIVE_CHAR;
        Serial.print (RECEIVE_CHAR);
        RECEIVE_BYTE_COUNT++;                     /* Increase byte counter */
      }
    }
  }
  while ((RECEIVE_CHAR != '\n') && (RECEIVE_CHAR != '\r') && (RECEIVE_BYTE_COUNT < 20));

  /* Not ENTER pressed and string not too long --> Save value */
  if ((RECEIVE_BYTE_COUNT > 0) && (RECEIVE_BYTE_COUNT < 20))                         
  {
    memset(NTP_SERVER_1, 0, sizeof(NTP_SERVER_1));/* Clear contents of NTP 1 server name */ 
    strcat(NTP_SERVER_1, INPUT_STRING);           /* Copy string */
  }
  Serial.println ();
  Serial.print ("  NTP 1 name set to \"");
  Serial.print (NTP_SERVER_1);
  Serial.println ("\"");   

  /* Step 10: Change actual NTP 2 server */
  delay (500);                                    /* 500 ms delay */
  while( Serial.available())                      /* Clear incoming serial buffer */
  { 
    RECEIVE_CHAR = Serial.read();                 /* Read char if available */
  }  
  Serial.print ("Actual NTP 2 name is \"");
  Serial.print (NTP_SERVER_2);
  Serial.print ("\", enter new value (<20 chars) or press <ENTER> to keep current name: ");
  memset(INPUT_STRING, 0, sizeof(INPUT_STRING));  /* Clear contents of input buffer */
  RECEIVE_BYTE_COUNT = 0;                         /* Reset byte counter */
  RECEIVE_CHAR = 32;                              /* Delete received char */
  do
  {
    if (Serial.available()>0)                     /* New character available ? */
    {
      RECEIVE_CHAR = Serial.read();               /* Read char if available */
      /* No CR and no LF received */
      if ((RECEIVE_CHAR != '\n') && (RECEIVE_CHAR != '\r'))
      {
        /* Save char */
        INPUT_STRING [RECEIVE_BYTE_COUNT] = RECEIVE_CHAR;
        Serial.print (RECEIVE_CHAR);
        RECEIVE_BYTE_COUNT++;                     /* Increase byte counter */
      }
    }
  }
  while ((RECEIVE_CHAR != '\n') && (RECEIVE_CHAR != '\r') && (RECEIVE_BYTE_COUNT < 20));

  /* Not ENTER pressed and string not too long --> Save value */
  if ((RECEIVE_BYTE_COUNT > 0) && (RECEIVE_BYTE_COUNT < 20))                         
  {
    memset(NTP_SERVER_2, 0, sizeof(NTP_SERVER_2));/* Clear contents of NTP 2 server name */ 
    strcat(NTP_SERVER_2, INPUT_STRING);           /* Copy string */
  }
  Serial.println ();
  Serial.print ("  NTP 2 name set to \"");
  Serial.print (NTP_SERVER_2);
  Serial.println ("\"");     

  /* Step 11: Change actual NTP 3 server */
  delay (500);                                    /* 500 ms delay */
  while( Serial.available())                      /* Clear incoming serial buffer */
  { 
    RECEIVE_CHAR = Serial.read();                 /* Read char if available */
  }  
  Serial.print ("Actual NTP 3 name is \"");
  Serial.print (NTP_SERVER_3);
  Serial.print ("\", enter new value (<20 chars) or press <ENTER> to keep current name: ");
  memset(INPUT_STRING, 0, sizeof(INPUT_STRING));  /* Clear contents of input buffer */
  RECEIVE_BYTE_COUNT = 0;                         /* Reset byte counter */
  RECEIVE_CHAR = 32;                              /* Delete received char */
  do
  {
    if (Serial.available()>0)                     /* New character available ? */
    {
      RECEIVE_CHAR = Serial.read();               /* Read char if available */
      /* No CR and no LF received */
      if ((RECEIVE_CHAR != '\n') && (RECEIVE_CHAR != '\r'))
      {
        /* Save char */
        INPUT_STRING [RECEIVE_BYTE_COUNT] = RECEIVE_CHAR;
        Serial.print (RECEIVE_CHAR);
        RECEIVE_BYTE_COUNT++;                     /* Increase byte counter */
      }
    }
  }
  while ((RECEIVE_CHAR != '\n') && (RECEIVE_CHAR != '\r') && (RECEIVE_BYTE_COUNT < 20));

  /* Not ENTER pressed and string not too long --> Save value */
  if ((RECEIVE_BYTE_COUNT > 0) && (RECEIVE_BYTE_COUNT < 20))                         
  {
    memset(NTP_SERVER_3, 0, sizeof(NTP_SERVER_3));/* Clear contents of NTP 3 server name */ 
    strcat(NTP_SERVER_3, INPUT_STRING);           /* Copy string */
  }
  Serial.println ();
  Serial.print ("  NTP 3 name set to \"");
  Serial.print (NTP_SERVER_3);
  Serial.println ("\""); 

  /* Step 12: Save changes ? */
  delay (500);                                    /* 500 ms delay */
  while( Serial.available())                      /* Clear incoming serial buffer */
  { 
    RECEIVE_CHAR = Serial.read();                 /* Read char if available */
  }  
  Serial.println ();
  Serial.print ("Save changes (y/n)? ");

  /* Wait for char */
  while( Serial.available() <= 0) 
  {
  }

  if ( Serial.readBytesUntil('\n', YES_NO_BUFFER, 2) > 0)
  { 
    Serial.println (YES_NO_BUFFER[0]);            /* Print answer */
    /* Save selected ? */
    if( ((String)YES_NO_BUFFER[0] == "y") || ((String)YES_NO_BUFFER[0] == "Y"))
    {
      Serial.println ("Saving selected - Saving data in Init.txt: ");
      SUB_WRITE_INIT_FILE ();                     /* Save values in Init.txt */
      SUB_READ_INIT_FILE ();                      /* Retrieve data from Init.txt */    
    }
    else                                          /* No saving ? */
    {
      Serial.println ("No saving selected - Retrieving data from Init.txt...");
      SUB_READ_INIT_FILE ();                      /* Retrieve data from Init.txt */ 
    }
  }
}


/****************************************************************************************/
/****************************************************************************************/
/* DEBUG ROUTINES                                                                       */
/****************************************************************************************/
/****************************************************************************************/


/****************************************************************************************/
/* Display all FLASH values                                                             */
/****************************************************************************************/
void SUB_DISPLAY_FLASH (void)
{
  byte FLASH_VALUE;                               /* Reading FLASH value */
  
  delay (1);                                      /* Proforma delay */
  #ifdef _DEBUG_
    Serial.println ("List FLASH memory:"); 
    FLASH_VALUE = EEPROM.read(0);
    Serial.print ("Memory location 0: Actual program (1-13 = Colours, 15-19=Storage positions, 20-29=Programs): ");
    Serial.println (FLASH_VALUE);
    FLASH_VALUE = EEPROM.read(1);
    Serial.print ("Memory location 1: Actual dimming color (0=no dimming, 9=max dimming): ");
    Serial.println (FLASH_VALUE);   
    FLASH_VALUE = EEPROM.read(2);
    Serial.print ("Memory location 2: Actual dimming program (0=no dimming, 4=max dimming): ");
    Serial.println (FLASH_VALUE); 
    FLASH_VALUE = EEPROM.read(3);
    Serial.print ("Memory location 3: Actual program speed (0=highest, 9=lowest): ");
    Serial.println (FLASH_VALUE);     
    FLASH_VALUE = EEPROM.read(4);
    Serial.print ("Memory location 4: Actual automatic mode (0=Off, 1=Auto): ");
    Serial.println (FLASH_VALUE); 
    FLASH_VALUE = EEPROM.read(5);
    Serial.print ("Memory location 5: Actual slave 1 automatic mode (0=Off, 1=Auto): ");
    Serial.println (FLASH_VALUE);   
    FLASH_VALUE = EEPROM.read(7);
    Serial.print ("Memory location 7: Actual slave 2 automatic mode (0=Off, 1=Auto): ");
    Serial.println (FLASH_VALUE);  
    FLASH_VALUE = EEPROM.read(9);
    Serial.print ("Memory location 9: Actual slave 3 automatic mode (0=Off, 1=Auto): ");
    Serial.println (FLASH_VALUE);              
    FLASH_VALUE = EEPROM.read(10);
    Serial.print ("Memory location 10: Colour storage position 1 - Red value: ");
    Serial.println (FLASH_VALUE); 
    FLASH_VALUE = EEPROM.read(11);
    Serial.print ("Memory location 11: Colour storage position 1 - Green value: ");
    Serial.println (FLASH_VALUE); 
    FLASH_VALUE = EEPROM.read(12);
    Serial.print ("Memory location 12: Colour storage position 1 - Blue value: ");
    Serial.println (FLASH_VALUE);
    FLASH_VALUE = EEPROM.read(13);
    Serial.print ("Memory location 13: Colour storage position 1 - Dimming (0=no dimming, 9=max dimming): ");
    Serial.println (FLASH_VALUE);
    FLASH_VALUE = EEPROM.read(20);
    Serial.print ("Memory location 20: Colour storage position 2 - Red value: ");
    Serial.println (FLASH_VALUE); 
    FLASH_VALUE = EEPROM.read(21);
    Serial.print ("Memory location 21: Colour storage position 2 - Green value: ");
    Serial.println (FLASH_VALUE); 
    FLASH_VALUE = EEPROM.read(22);
    Serial.print ("Memory location 22: Colour storage position 2 - Blue value: ");
    Serial.println (FLASH_VALUE);
    FLASH_VALUE = EEPROM.read(23);
    Serial.print ("Memory location 23: Colour storage position 2 - Dimming (0=no dimming, 9=max dimming): ");
    Serial.println (FLASH_VALUE);
    FLASH_VALUE = EEPROM.read(30);
    Serial.print ("Memory location 30: Colour storage position 3 - Red value: ");
    Serial.println (FLASH_VALUE); 
    FLASH_VALUE = EEPROM.read(31);
    Serial.print ("Memory location 31: Colour storage position 3 - Green value: ");
    Serial.println (FLASH_VALUE); 
    FLASH_VALUE = EEPROM.read(32);
    Serial.print ("Memory location 32: Colour storage position 3 - Blue value: ");
    Serial.println (FLASH_VALUE);
    FLASH_VALUE = EEPROM.read(33);
    Serial.print ("Memory location 33: Colour storage position 3 - Dimming (0=no dimming, 9=max dimming): ");
    Serial.println (FLASH_VALUE);
    FLASH_VALUE = EEPROM.read(40);
    Serial.print ("Memory location 40: Colour storage position 4 - Red value: ");
    Serial.println (FLASH_VALUE); 
    FLASH_VALUE = EEPROM.read(41);
    Serial.print ("Memory location 41: Colour storage position 4 - Green value: ");
    Serial.println (FLASH_VALUE); 
    FLASH_VALUE = EEPROM.read(42);
    Serial.print ("Memory location 42: Colour storage position 4 - Blue value: ");
    Serial.println (FLASH_VALUE);
    FLASH_VALUE = EEPROM.read(43);
    Serial.print ("Memory location 43: Colour storage position 4 - Dimming (0=no dimming, 9=max dimming): ");
    Serial.println (FLASH_VALUE);    
    FLASH_VALUE = EEPROM.read(50);
    Serial.print ("Memory location 50: Colour storage position 5 - Red value: ");
    Serial.println (FLASH_VALUE); 
    FLASH_VALUE = EEPROM.read(51);
    Serial.print ("Memory location 51: Colour storage position 5 - Green value: ");
    Serial.println (FLASH_VALUE); 
    FLASH_VALUE = EEPROM.read(52);
    Serial.print ("Memory location 52: Colour storage position 5 - Blue value: ");
    Serial.println (FLASH_VALUE);
    FLASH_VALUE = EEPROM.read(53);
    Serial.print ("Memory location 53: Colour storage position 5 - Dimming (0=no dimming, 9=max dimming): ");
    Serial.println (FLASH_VALUE);
    Serial.print ("Memory location 60: Remote update flag: "); 
    Serial.print (EEPROM.read(60));
    Serial.println (" (Normal boot = 0, Remote update = 1)");
    Serial.print ("Memory location 61-62: NTP Update time: ");
    FLASH_VALUE = EEPROM.read(61);
    if (FLASH_VALUE < 10)
      Serial.print ("0");
    Serial.print (FLASH_VALUE);
    Serial.print (":"); 
    FLASH_VALUE = EEPROM.read(62);
    if (FLASH_VALUE < 10)
      Serial.print ("0");
    Serial.println (FLASH_VALUE);    
    Serial.print ("Memory location 63: NTP time zone: "); 
    FLASH_VALUE = EEPROM.read(63);
    switch (FLASH_VALUE)                          /* Call sub menues after choosing a menu item */
    {
      case 0:                                     /* 0 = -5 hours */ 
        Serial.println ("-5 hours");          
        break;
      case 1:                                     /* 1 = -4 hours */ 
        Serial.println ("-4 hours");          
        break;
      case 2:                                     /* 2 = -3 hours */ 
        Serial.println ("-3 hours");          
        break;
      case 3:                                     /* 3 = -2 hours */ 
        Serial.println ("-2 hours");          
        break;
      case 4:                                     /* 4 = -1 hour */ 
        Serial.println ("-1 hour");          
        break;
      case 5:                                     /* 5 = 0 hours */ 
        Serial.println ("0 hours");          
        break;
      case 6:                                     /* 6 = +1 hour */ 
        Serial.println ("+1 hour");          
        break;  
      case 7:                                     /* 7 = +2 hours */ 
        Serial.println ("+2 hours");          
        break;
      case 8:                                     /* 8 = +3 hours */ 
        Serial.println ("+3 hours");          
        break;
      case 9:                                     /* 9 = +4 hours */ 
        Serial.println ("+4 hours");          
        break;
      case 10:                                    /* 10 = +5 hours */ 
        Serial.println ("+5 hours");          
        break;
    }   
    FLASH_VALUE = EEPROM.read(64);
    if (FLASH_VALUE == 0)
    {
       Serial.println ("Memory location 64: NTP automatic summertime switching: Disabled");  
    }
    else
    {
      Serial.println ("Memory location 64: NTP automatic summertime switching: Enabled"); 
    }                      
  # endif  
}


/****************************************************************************************/
/* Subroutine to send clock and date via serial interface to terminal                   */
/****************************************************************************************/
void SUB_CLOCK_PRINT()                             
{
  delay (1);                                      /* Proforma delay */
  #ifdef _DEBUG_
    Serial.print ("Time: ");                      /* Display time header */
    Serial.print(TIME_STRING);                    /* Display time info */
    Serial.print (":00, Date: ");                 /* Display date header */
    Serial.println (DATE_STRING);                 /* DIsplay date string */
    Serial.print("Signal strength (RSSI): ");     /* Display WIFI signal strength */
    Serial.print(WiFi.RSSI());
    Serial.println(" dBm");
    Serial.println ();
  # endif 
}


/****************************************************************************************/
/****************************************************************************************/
/* WATCHDOG ROUTINES                                                                    */
/****************************************************************************************/
/****************************************************************************************/


/****************************************************************************************/
/* Subroutine to trigger watchdog                                                       */
/****************************************************************************************/
void SUB_TRIGGER_WATCHDOG (void)
{
  if (WD_FLAG == true)                            /* WD flag true = High ? */
  { 
    digitalWrite (WD_TRIGGER, LOW);               /* Set WD trigger output to Low */
    WD_FLAG = false;                              /* WD flag false = Low */
  }
  else
  {
    digitalWrite (WD_TRIGGER, HIGH);              /* Set WD trigger output to High */
    WD_FLAG = true;                               /* WD flag true = High */
  }
}


/****************************************************************************************/
/* Subroutine wait for hardware reset via watchog reset                                 */
/****************************************************************************************/
void SUB_WAIT_FOR_WATCHDOG (void)
{
  #ifdef _DEBUG_
    Serial.print("  Resetting ESP device via watchdog!!!");
  # endif
  SUB_TRIGGER_WATCHDOG ();                        /* Trigger watchdog in case not triggered before */
  /* Endless loop until watchdog reset */
  do                                              /* Endless loop until watchdog reset */
  {
    delay (12000);                                /* Wait 12 Seconds */
    SUB_TRIGGER_WATCHDOG ();                      /* Watchdog trigger just to be on the safe side */
  }
  while (1);
}


/****************************************************************************************/
/****************************************************************************************/
/* BOOTLOADER ROUTINES                                                                  */
/****************************************************************************************/
/****************************************************************************************/

/****************************************************************************************/
/* Subroutine to initialize WLAN bootloader and wait for upload                         */
/* Port defaults to 3232 (ArduinoOTA.setPort(3232))                                     */
/* Hostname defaults to esp3232-[MAC] (ArduinoOTA.setHostname("myesp32"))               */
/* No authentication by default (ArduinoOTA.setPassword("admin"))                       */
/* Password can be set with it's md5 value as well                                      */
/*      MD5(admin) = 21232f297a57a5a743894a0e4a801fc3                                   */
/*      ArduinoOTA.setPasswordHash("21232f297a57a5a743894a0e4a801fc3")                  */
/****************************************************************************************/
void SUB_WLAN_BOOTLOADER (void)
{
  byte KEY_VALUE;                                 /* Key Value of pressed button */
  unsigned long AWAKE_END;                        /* Variable for determing display timeout */
  unsigned long CURRENT_TIME;                     /* Variable for determing display timeout */
  boolean BLINK_FLAG;                             /* True: Led is on, false = LED is off */
   
  #ifdef _DEBUG_
    Serial.println();
    Serial.print ("Initializing bootloader: ");
  # endif 

  /* Next boot = normal boot */
  EEPROM.write(60, 0);                            /* Set REMOTE UPLOAD FLAG: for 0=Normal  */
  EEPROM.commit();                                /* Write to FLASH memory */
  
  ArduinoOTA
    .onStart([]()                                 /* On start */
    {
      String type;
      if (ArduinoOTA.getCommand() == U_FLASH)     /* Sketch upload ? */
        type = "sketch";
      else                                        /* U_SPIFFS selected ? */
        type = "filesystem";
      /* NOTE: if updating SPIFFS this would be the place to unmount SPIFFS using SPIFFS.end() */
      Serial.println("Start updating " + type);
    })
    
    .onEnd([]()                                   /* Finished upload */ 
    {
      Serial.println("\nEnd");
    })

    /* On progress */
    .onProgress([](unsigned int progress, unsigned int total) 
    {
      Serial.printf("Progress: %u%%\r", (progress / (total / 100)));
    })
    
    .onError([](ota_error_t error)                /* On error */
    {
      Serial.printf("Error[%u]: ", error);
      if (error == OTA_AUTH_ERROR) Serial.println("Auth Failed");
      else if (error == OTA_BEGIN_ERROR) Serial.println("Begin Failed");
      else if (error == OTA_CONNECT_ERROR) Serial.println("Connect Failed");
      else if (error == OTA_RECEIVE_ERROR) Serial.println("Receive Failed");
      else if (error == OTA_END_ERROR) Serial.println("End Failed");
    });

  ArduinoOTA.begin();                             /* Start OTA */

  /* Endless loop to wait for request from Arduino IDE */
  CURRENT_TIME = millis();                        /* Get actual time value */
  AWAKE_END = CURRENT_TIME + 500;                 /* Set new blink value to 0,5 sec */
  SUB_DMX_SEND (0, 255, 0);                       /* Switch all pixels to green color */
  BLINK_FLAG = true;                              /* Set blink flag true */
  do
  {
    ArduinoOTA.handle();                          /* OTA handler */
    /* Check if display should be switched off after timeout */
    CURRENT_TIME = millis();                      /* Get actual time value */
    /* Check if blinking reached */
    if(CURRENT_TIME > AWAKE_END) 
    {
      if (BLINK_FLAG == true)                     /* Switch LED off ? */
      {
        SUB_DMX_SEND (0, 0, 0);                   /* Switch off all pixels */
        BLINK_FLAG = false;                       /* Set blink flag false */
      }
      else                                        /* Switch LED on ? */
      {
        SUB_DMX_SEND (0, 255, 0);                 /* Switch all pixels to green color */
        BLINK_FLAG = true;                        /* Set blink flag true */
      }
      AWAKE_END = CURRENT_TIME + 500;             /* Set new blink value to 0,5 sec */
    }
  }
  while (true);
}



/****************************************************************************************/
/****************************************************************************************/
/* DMX routines                                                                         */
/****************************************************************************************/
/****************************************************************************************/

/****************************************************************************************/
/* Subroutine send DMX channels                                                         */
/****************************************************************************************/
void SUB_DMX_SEND(byte color_1, byte color_2, byte color_3)
{
  xSemaphoreTake( ESP32DMX.lxDataLock, portMAX_DELAY );
  if (SLAVE_1_MODE == 0)                          /* Slave mode 1 is zero = off ? */
  {
     ESP32DMX.setSlot(1 , 0);                     /* Update DMX channel 1 with zero value */ 
     ESP32DMX.setSlot(2 , 0);                     /* Update DMX channel 2 with zero value */
     ESP32DMX.setSlot(3 , 0);                     /* Update DMX channel 3 with zero value */    
  }
  else                                            /* Slave mode 1 is not zero = Auto ? */
  {
    ESP32DMX.setSlot(1 , color_1);                /* Update DMX channel 1 with red value */ 
    ESP32DMX.setSlot(2 , color_2);                /* Update DMX channel 2 with green value */
    ESP32DMX.setSlot(3 , color_3);                /* Update DMX channel 3 with blue value */
  }
  if (SLAVE_2_MODE == 0)                          /* Slave mode 1 is zero = off ? */
  {
     ESP32DMX.setSlot(4 , 0);                     /* Update DMX channel 4 with zero value */ 
     ESP32DMX.setSlot(5 , 0);                     /* Update DMX channel 5 with zero value */
     ESP32DMX.setSlot(6 , 0);                     /* Update DMX channel 6 with zero value */    
  }
  else                                            /* Slave mode 1 is not zero = Auto ? */
  {
    ESP32DMX.setSlot(4 , color_1);                /* Update DMX channel 4 with red value */ 
    ESP32DMX.setSlot(5 , color_2);                /* Update DMX channel 5 with green value */
    ESP32DMX.setSlot(6 , color_3);                /* Update DMX channel 6 with blue value */    
  }
  if (SLAVE_3_MODE == 0)                          /* Slave mode 2 is zero = off ? */
  {
     ESP32DMX.setSlot(7 , 0);                     /* Update DMX channel 7 with zero value */ 
     ESP32DMX.setSlot(8 , 0);                     /* Update DMX channel 8 with zero value */
     ESP32DMX.setSlot(9 , 0);                     /* Update DMX channel 9 with zero value */    
  }
  else                                            /* Slave mode 1 is not zero = Auto ? */
  {
    ESP32DMX.setSlot(7 , color_1);                /* Update DMX channel 7 with red value */ 
    ESP32DMX.setSlot(8 , color_2);                /* Update DMX channel 8 with green value */
    ESP32DMX.setSlot(9 , color_3);                /* Update DMX channel 9 with blue value */    
  }  
  xSemaphoreGive( ESP32DMX.lxDataLock );
}
    
    
    
    
    
    
    
