/****************************************************************************************/
/* Wifi.ino                                                                             */
/* Collection of subroutines for Wifi and NTP                                           */
/* 1. WIFI routines:                                                                    */
/* SUB_WLAN_CONNECT:       Subroutine to connect to WIFI                                */
/* SUB_CHECK_WIFI_CONNECT: Subroutine to check if WLAN connection is still on           */
/*                                                                                      */
/* 2. NTP routines:                                                                     */
/* SUB_NTP_RECEIVE:           Subroutine to receive NTP time                            */
/* SUB_NTP_ADJUST_SUMMERTIME: Subroutine to adjust received time to timezone            */          
/****************************************************************************************/


/****************************************************************************************/
/****************************************************************************************/
/* WIFI routines                                                                        */
/****************************************************************************************/
/****************************************************************************************/


/****************************************************************************************/
/* Subroutine to connect to WIFI                                                        */
/****************************************************************************************/
void SUB_WLAN_CONNECT (void)
{
  uint16_t WLAN_COUNTER;                          /* Wifi counter to check successfull connection */
  int      WLAN_STATUS;                           /* Actual Wifi status */
  boolean  PING_STATUS;                           /* True = Responding to ping, False = No response to ping */  
  
  WLAN_CONNECT = true;                            /* Set Wifi connect = true */
  do                                              /* Repeat until Wifi is connected */
  {
    #ifdef _DEBUG_
      Serial.println("Starting to connect to WLAN ...");
      Serial.print("  Disconnecting from Wifi: ");  
    # endif 
    /* Step A: Disconnect from WLAN if still connected */
    if (WiFi.status() == WL_CONNECTED)            /* Wifi connected? */
    {
      WiFi.disconnect();                          /* Disconnect from Wifi */
      #ifdef _DEBUG_
        Serial.println("Disconnected");
      # endif 
    }  
    else                                          /* Wifi not connected? */
    {
      #ifdef _DEBUG_
        Serial.println("No connection");
      # endif 
    }   
    
    /* Step B: Set hostname and setting fixed IP-Address */
    #ifdef _DEBUG_
      Serial.print("  Setting fixed IP address: "); 
    # endif
    /* Set fixed IP address failes ? */
    /* DNS servers must be set so that NTP can work! */
    if (!WiFi.config(FIXED_IP, GATEWAY, SUBNET, DNS_1, DNS_2)) 
    {
      #ifdef _DEBUG_
        Serial.println("Failed");
      # endif
      WLAN_CONNECT = false;                       /* Set Wifi connect = true = Error */
    }  
    else                                          /* Set fixed IP address OK ? */
    {
      #ifdef _DEBUG_
        Serial.println("Done");
      # endif
    }

    /* Step C: Disable Wifi Sleep and config Station mode */
    WiFi.setSleep(false);                         /* Switch sleep mode off */ 
    WiFi.setHostname(HOSTNAME);                   /* Set hostname for ESP */
    WiFi.mode (WIFI_STA);                         /* Configure station mode */
    
    if (WLAN_CONNECT == true)                     /* No error yet ? */
    {
      /* Step D: Connecting to Wifi Network */
      #ifdef _DEBUG_
        Serial.print("  Connecting to WLAN ");
        Serial.print(WLAN_SSID); 
        Serial.print(": "); 
      # endif
      WiFi.begin(WLAN_SSID, WLAN_PWD);            /* Connecting to WLAN */
      
      WLAN_COUNTER = 0;                           /* Set Wifi counter 0 */
      /* Wifi connected or 20 trials done ? */
      while ((WiFi.status() != WL_CONNECTED) && (WLAN_COUNTER < 20)) 
      {
        delay (100);                              /* Wait hundret miliseconds */
        WLAN_COUNTER++;                           /* Increment Wifi counter */
      }

      if (WLAN_COUNTER >= 20)                     /* Error connecting */
      {
        #ifdef _DEBUG_
          Serial.println("Failed");
        # endif
        WiFi.disconnect();                        /* Disconnect from Wifi */
        delay (1000);                             /* Waiting one second */
        #ifdef _DEBUG_
          Serial.print("  Reconnecting to WLAN ");
          Serial.print(WLAN_SSID); 
          Serial.print(": "); 
        # endif  
        WiFi.begin(WLAN_SSID, WLAN_PWD);          /* Connecting to WLAN */
        WLAN_COUNTER = 0;                         /* Set Wifi counter 0 */
        /* Wifi connected or 20 trials done ? */
        while ((WiFi.status() != WL_CONNECTED) && (WLAN_COUNTER < 20)) 
        {
          delay (100);                            /* Wait hundret miliseconds */
          WLAN_COUNTER++;                         /* Increment Wifi counter */
        }
      }
      
      if (WLAN_COUNTER >= 20)                     /* Error connecting */
      {
        #ifdef _DEBUG_
          Serial.println("Failed");
        # endif       
      }
      else                                        /* Successfull connection ? --> Start ping */
      {
        /* Display Wifi information */
        #ifdef _DEBUG_
          Serial.println("Done"); 
          Serial.print ("  Ping server for connectivity check: ");
          Serial.print (GATEWAY);
          Serial.print(": "); 
        # endif
        PING_STATUS = Ping.ping(GATEWAY, 4);      /* Ping to IP adress */
        if (PING_STATUS == true)                  /* Sucessful ping? */
        {
          #ifdef _DEBUG_
            Serial.println("Success");
          # endif
        }
        else                                      /* Unsucessful ping? */
        {
          #ifdef _DEBUG_
            Serial.println("Failed");
          # endif
          WLAN_COUNTER = 21;                      /* Set Wifi counter to error position */
        }  
      }
      
      if (WLAN_COUNTER >= 20)                     /* Error connecting or pinging ? */
      { 
        WiFi.disconnect();                        /* Wifo disconnect */
        WiFi.mode (WIFI_OFF);                     /* Switch Wifi off */
        SUB_DMX_SEND (255, 0, 0);                 /* Switch all pixels to red color */
        delay (2000);                             /* 2 seconds delay */
        SUB_DMX_SEND (0, 0, 0);                   /* Switch off all pixels */
        delay (2000);                             /* 2 seconds delay */
        SUB_WAIT_FOR_WATCHDOG ();                 /* Wait for watchdog reset */
      }
      else                                        /* Wifi connected sucessfully and ping OK ? */
      {
        #ifdef _DEBUG_
          /* Display Wifi information */
          Serial.print("    IP address: ");
          Serial.println(WiFi.localIP());
          Serial.print("    ESP Mac Address: ");
          Serial.println(WiFi.macAddress());
          Serial.print("    Subnet Mask: ");
          Serial.println(WiFi.subnetMask());
          Serial.print("    Gateway IP: ");
          Serial.println(WiFi.gatewayIP()); 
          Serial.print("    Hostname: ");
          Serial.println(WiFi.getHostname());         
          Serial.print("    Signal strength (RSSI): ");
          Serial.print(WiFi.RSSI());
          Serial.println(" dBm");
          Serial.print("    WLAN status: ");
          WLAN_STATUS = WiFi.status();
          switch (WLAN_STATUS)
          {
            case 255:
              Serial.println("WL_NO_SHIELD");
              break;
            case 0:
              Serial.println("WL_IDLE_STATUS");
              break;
            case 1:
              Serial.println("WL_NO_SSID_AVAIL");
              break;
            case 2:
              Serial.println("WL_SCAN_COMPLETED");
              break;
            case 3:
              Serial.println("WL_CONNECTED");
              break;
            case 4:
              Serial.println("WL_CONNECT_FAILED");
              break;
            case 5:
              Serial.println("WL_CONNECTION_LOST");
              break;
            case 6:
              Serial.println("WL_DISCONNECTED");
              break;
            default:
              Serial.println("_UNKNOWN");
              break;
          }          
          Serial.print("    Use this URL: http://");
          Serial.print(WiFi.localIP());
          Serial.println("/main");          
        # endif 
        WLAN_CONNECT = true;                      /* Set WLAN connect flag to true */
      }
    }       
  }
  while (WLAN_CONNECT == false);
}


/****************************************************************************************/
/* Subroutine to check if WLAN connection is still on                                   */
/****************************************************************************************/
void SUB_CHECK_WIFI_CONNECT (void)
{
  boolean  PING_STATUS;                           /* True = Responding to ping, False = No response to ping */

  #ifdef _DEBUG_
    Serial.print ("Checking WLAN connection: ");
  # endif 

  if (WiFi.status() == WL_CONNECTED)              /* Wifi connected? */
  {
    PING_STATUS = Ping.ping(GATEWAY, 4);          /* Ping to IP adress */
    if (PING_STATUS == true)                      /* Sucessful ping? */
    {
      #ifdef _DEBUG_
        Serial.println("Ok");
      # endif
    }
    else                                          /* Unsucessful ping? */
    {
      #ifdef _DEBUG_
        Serial.println("Error");
      # endif
      SUB_WLAN_CONNECT ();                        /* Call subroutine to try to connect to WLAN */
    }
  }
  else                                            /* Wifi not connected? */
  {
    #ifdef _DEBUG_
      Serial.println("Error");
    # endif
    SUB_WLAN_CONNECT ();                          /* Call subroutine to try to connect to WLAN */
  }
  #ifdef _DEBUG_
    Serial.println();
  # endif
}


/****************************************************************************************/
/****************************************************************************************/
/* NTP routines                                                                         */
/****************************************************************************************/
/****************************************************************************************/


/****************************************************************************************/
/* NTP-Clock: Subroutine to receive NTP time                                            */
/*                                                                                      */
/* %a Abbreviated weekday name                                                          */
/* %A Full weekday name                                                                 */
/* %b Abbreviated month name                                                            */
/* %B Full month name                                                                   */
/* %c Date and time representation for your locale                                      */
/* %d Day of month as a decimal number (01-31)                                          */
/* %H Hour in 24-hour format (00-23)                                                    */
/* %I Hour in 12-hour format (01-12)                                                    */
/* %j Day of year as decimal number (001-366)                                           */
/* %m Month as decimal number (01-12)                                                   */
/* %M Minute as decimal number (00-59)                                                  */
/* %p Current locale's A.M./P.M. indicator for 12-hour clock                            */
/* %S Second as decimal number (00-59)                                                  */
/* %U Week of year as decimal number,  Sunday as first day of week (00-51)              */
/* %w Weekday as decimal number (0-6; Sunday is 0)                                      */
/* %W Week of year as decimal number, Monday as first day of week (00-51)               */
/* %x Date representation for current locale                                            */
/* %X Time representation for current locale                                            */
/* %y Year without century, as decimal number (00-99)                                   */
/* %Y Year with century, as decimal number                                              */
/* %z %Z Time-zone name or abbreviation, (no characters if time zone is unknown)        */
/* %% Percent sign                                                                      */
/* You can include text literals (such as spaces and colons) to make a neater display   */
/* or for padding between adjoining columns.                                            */
/* You can suppress the display of leading zeroes  by using the "#" character           */
/* (%#d, %#H, %#I, %#j, %#m, %#M, %#S, %#U, %#w, %#W, %#y, %#Y)                         */
/*                                                                                      */
/* Variables of structure tm:                                                           */
/*  int tm_sec                                                                          */
/*  int tm_min                                                                          */
/*  int tm_hour                                                                         */
/*  int tm_mday                                                                         */
/*  int tm_mon                                                                          */
/*  int tm_year                                                                         */
/*  int tm_wday                                                                         */
/*  int tm_yday                                                                         */
/*  int tm_isdst                                                                        */
/****************************************************************************************/
void SUB_NTP_RECEIVE()
{
  NTP_CONNECT = false;                            /* Set NTP connect = true */

  #ifdef _DEBUG_
    Serial.println("Retrieving NTP time and date ...");
    Serial.print("  Receive first NTP Sync: ");  
  # endif   

  DAYLIGHT_OFFSET_SEC = 0;                        /* Set to winter time */
  /* Gets the actual time via NTP server */
  configTime(GMT_OFFSET_SEC, DAYLIGHT_OFFSET_SEC, NTP_SERVER_1, NTP_SERVER_2, NTP_SERVER_3);  
  if(!getLocalTime(&TIMEINFO))                    /* Reception of NTP time not OK ? */
  {
    #ifdef _DEBUG_
      Serial.println(" Failed");
    # endif
    return;                                       /* Return from subroutine */
  }
  else                                            /* Reception of NTP time OK ? */
  {
    #ifdef _DEBUG_
      Serial.println(" Done"); 
      Serial.print ("    Time: ");
      Serial.print(&TIMEINFO, "%H:%M:%S");
      Serial.print (", Date: ");
      Serial.println (&TIMEINFO, "%d.%m.%Y");
    # endif
    ACT_YEAR = TIMEINFO.tm_year +1900;            /* Extract actual year */
    ACT_MONTH = TIMEINFO.tm_mon + 1;              /* Extract actual month */
    ACT_DAY = TIMEINFO.tm_mday;                   /* Extract actual day */     
    /* Summertime detected and NTP summertime flag = true = ummertime */
    if((SUB_NTP_ADJUST_SUMMERTIME() == true) && (NTP_SUMMERTIME == true))    
    {
      DAYLIGHT_OFFSET_SEC = 3600;                 /* Set to summertime */ 
      #ifdef _DEBUG_
        Serial.println ("    Summertime detected");
        Serial.print("  Receive second NTP Sync: ");  
      # endif  
  
      /* Gets the actual time via NTP server */
      configTime(GMT_OFFSET_SEC, DAYLIGHT_OFFSET_SEC, NTP_SERVER_1, NTP_SERVER_2, NTP_SERVER_3);  
      if(!getLocalTime(&TIMEINFO))                /* Reception of NTP time not OK ? */
      {
        #ifdef _DEBUG_
          Serial.println(" Failed");
        # endif
        return;                                   /* Return from subroutine */
      }
      else                                        /* Reception of NTP time OK ? */
      {
        #ifdef _DEBUG_
          Serial.println(" Done"); 
          Serial.print ("    Time: ");
          Serial.print(&TIMEINFO, "%H:%M:%S");
          Serial.print (", Date: ");
          Serial.println (&TIMEINFO, "%d.%m.%Y");
        # endif  
        NTP_CONNECT = true;                       /* Set NTP connect = true */
        return;                                   /* Return from subroutine */ 
      } 
    }
    else                                          /* Wintertime detected or no automatic detection of summertime ? */
    {
      #ifdef _DEBUG_
        if (NTP_SUMMERTIME == false)              /* No automatic summertime detection ? */
        {
          Serial.print ("    Automatic summertie switching disabled");       
        }
        else                                      /* Automatic summertime detection ? */
        {
          Serial.print ("    Wintertime detected");
        } 
      # endif
      DAYLIGHT_OFFSET_SEC = 0;                    /* Set to wintertime */ 
      NTP_CONNECT = true;                         /* Set NTP connect = true */
      return;                                     /* Return from subroutine */ 
    }
  }
}


/****************************************************************************************/
/* NTP-Clock: Subroutine to adjust received time to timezone                            */
/* Returns tue if summertine and false if wintertime                                    */            
/****************************************************************************************/
boolean SUB_NTP_ADJUST_SUMMERTIME()
{
  int BEGIN_DST_DATE;                             /* Begin date */
  int BEGIN_DST_MONTH;                            /* Begin month */
  int END_DST_DATE;                               /* End date */
  int END_DST_MONTH;                              /* End month */

  /* Determine last Sunday of March */
  BEGIN_DST_DATE = (31 - (5* ACT_YEAR /4 + 4) % 7); 
  BEGIN_DST_MONTH = 3;                            /* Begin month is March */
  #ifdef _DEBUG_
    Serial.print ("    Actual summer time begin date: ");
    Serial.print (BEGIN_DST_DATE);
    Serial.print (".");
    Serial.println (BEGIN_DST_MONTH);  
  # endif
  /* Determine last Sunday of October */
  END_DST_DATE = (31 - (5 * ACT_YEAR /4 + 1) % 7);   
  END_DST_MONTH = 10;                             /* Begin month is October */
  #ifdef _DEBUG_
    Serial.print ("    Actual summer time end date: ");
    Serial.print (END_DST_DATE);
    Serial.print (".");
    Serial.println (END_DST_MONTH);  
  # endif    
  /* Check for summer time */
  if (((ACT_MONTH > BEGIN_DST_MONTH) && (ACT_MONTH < END_DST_MONTH))
      || ((ACT_MONTH == BEGIN_DST_MONTH) && (ACT_DAY >= BEGIN_DST_DATE))
      || ((ACT_MONTH == END_DST_MONTH) && (ACT_DAY <= END_DST_DATE)))
  {
    return true;                                  /* Return true = Summer time */
  }
  else                                            /* Winter time ? */
  {
    return false;                                 /* Return true = Summer time */
  }
} 
