/****************************************************************************************/
/* DMX zu Neo Pixel converter garden lightening                                         */
/* Version: 2.0                                                                         */
/* Date: 25.06.2020                                                                     */
/* Compiled with Arduino 1.8.12                                                         */
/* Arduino board: Arduino Leonardo                                                      */
/*                                                                                      */
/*                                                                                      */
/* Pin asignments:                                                                      */
/*    Neopixel Control Pin = 6                                                          */
/*    DMX Receive Pin = RXD                                                             */
/*    Slave configuration pins: D0 = Pin 2                                              */
/*    Slave configuration pins: D1 = Pin 4                                              */
/*    Slave configuration pins: D2 = Pin 7                                              */
/*                  000 = Slave 1 (DMX chaneels 1-3                                     */
/*                  100 = Slave 2 (DMX chaneels 4-6                                     */
/*                  010 = Slave 3 (DMX chaneels 7-9 (Default)                           */
/*                                                                                      */
/* Libraries: Fast LED: https://github.com/FastLED/FastLED                              */
/*            DMX Serial https://github.com/mathertel/DMXSerial                         */
/*                                                                                      */
/* DMX channel 1 = Green                                                                */
/* DMX channel 2 = Red                                                                  */
/* DMX channel 3 = Blue                                                                 */
/*                                                                                      */
/****************************************************************************************/

#include <FastLED.h>                              /* Include Neo Pixel fast LED library */
#include <DMXSerial.h>                            /* Include DMX serial library */

/****************************************************************************************/
/* General NEO Pixel settings                                                           */
/****************************************************************************************/
#define NEO_PIXEL_PIN 6                           /* Set Neo Pixel control pin to pin 6 */               
#define NUM_NEO_PIXELS 48                         /* Set number of Neo Pixels */
#define DMXSTART 1                                /* Define DMX start channel */
#define DMXLENGTH 9                               /* Define number of DMX channels */
CRGB leds[NUM_NEO_PIXELS];                        /* Define the array of leds */

/****************************************************************************************/
/* Other pin settings                                                                   */
/****************************************************************************************/
#define LED_PIN 17                                /* Define Led pin = 17 */
#define D0_PIN 2                                  /* Define slave pin D0 = 2 */
#define D1_PIN 4                                  /* Define slave pin D1 = 4 */
#define D2_PIN 7                                  /* Define slave pin D2 = 7 */

/****************************************************************************************/
/* Variables                                                                            */
/****************************************************************************************/
byte LOOP_COUNTER;                                /* Loop vale */
byte GREEN_VALUE;                                 /* Green value */
byte RED_VALUE;                                   /* Red value */
byte BLUE_VALUE;                                  /* Blue value */
byte OLD_GREEN_VALUE;                             /* OLD green value */
byte OLD_RED_VALUE;                               /* OLD red value */
byte OLD_BLUE_VALUE;                              /* OLD blue value */
byte SLAVE_MODE;                                  /* 1 = Slave mode 1, 4 = Slave mode 2, 7 = Slave mode 3 */
unsigned int LAST_PACKET_RECEIVED;                /* Time between two packets */

/****************************************************************************************/
/* Setup procedure                                                                      */
/****************************************************************************************/
void setup()
{
  /* Initialise DMV library */                    
  DMXSerial.init(DMXProbe);                       /* Init DMX library in probe mode */
  DMXSerial.maxChannel(DMXLENGTH);                /* Set maximum DMX channels */

  /* Initialise Pins */  
  pinMode(LED_BUILTIN, OUTPUT);                   /* Initialize internal LED as output */
  digitalWrite(LED_PIN, HIGH);                    /* Switch LED off */
  pinMode(D0_PIN, INPUT_PULLUP);                  /* Initialize D0 as input with Pullup */
  pinMode(D1_PIN, INPUT_PULLUP);                  /* Initialize D1 as input with Pullup */ 
  pinMode(D2_PIN, INPUT_PULLUP);                  /* Initialize D2 as input with Pullup */
      
  /* Iinitializes the NeoPixel library. */
  FastLED.addLeds<WS2812B, NEO_PIXEL_PIN, RGB>(leds, NUM_NEO_PIXELS);                                  
  
  delay (500);                                    /* Wait 500 ms */ 

  /* Clear all pixels */
  SUB_NEO_SEND (0, 0, 0);                         /* Call subroutine to clear all pixel */

  /* Initialise OLD values */
  OLD_GREEN_VALUE = 0;                            /* Set OLD green value to zero */
  OLD_RED_VALUE = 0;                              /* Set OLD red value to zero */
  OLD_BLUE_VALUE = 0;                             /* Set OLD blue value to zero */
  LAST_PACKET_RECEIVED = 0;                       /* Initialise last package counter */

  /* Check for Slave mode */
  SLAVE_MODE = 7;                                 /* Set slave mode to default = slave mode 3 */
  /* Slave 1 selected ? */
  if ((digitalRead(D0_PIN)==0) && (digitalRead(D1_PIN)==0)&& (digitalRead(D2_PIN)==0))
  {
    SLAVE_MODE = 1;                               /* Set slave mode to 1 */    
  }
  /* Slave 2 selected ? */
  if ((digitalRead(D0_PIN)==1) && (digitalRead(D1_PIN)==0)&& (digitalRead(D2_PIN)==0))
  {
    SLAVE_MODE = 4;                               /* Set slave mode to 2 */    
  }  
}


/****************************************************************************************/
/* Main Loop                                                                            */
/****************************************************************************************/
void loop()
{
  LAST_PACKET_RECEIVED ++;                        /* Increase last package receive counter */
  /* Ceck for received DMX frames */
  if (DMXSerial.receive() == true)                /* DMX frames received ? */
  {
    digitalWrite(LED_PIN, LOW);                   /* Switch internal LED on */
    GREEN_VALUE = DMXSerial.read(SLAVE_MODE);     /* Read green value from DMX channel */
    RED_VALUE = DMXSerial.read(SLAVE_MODE+1);     /* Read red value from DMX channel 5 */
    BLUE_VALUE = DMXSerial.read(SLAVE_MODE+2);    /* Read blue value from DMX channel 6 */
    LAST_PACKET_RECEIVED = 0;                     /* Reset last package counter */
    /* Value has changed ? */
    if ((GREEN_VALUE != OLD_GREEN_VALUE) || (RED_VALUE != OLD_RED_VALUE) || (BLUE_VALUE != OLD_BLUE_VALUE))
    {
      /* Call subroutine to adjust to received color */
      SUB_NEO_SEND (GREEN_VALUE , RED_VALUE, BLUE_VALUE);                         
      OLD_GREEN_VALUE = GREEN_VALUE;              /* Set OLD value green */
      OLD_RED_VALUE = RED_VALUE;                  /* Set OLD value red */
      OLD_BLUE_VALUE = BLUE_VALUE;                /* Set OLD value blue */
    }
  }
  else                                            /* DMX frames received ? */
  {
    if (LAST_PACKET_RECEIVED > 50)                /* Timeout after some seconds */               
    {
      digitalWrite(LED_PIN, HIGH);                /* Switch internal LED off */
      SUB_NEO_SEND (50, 0, 0);                    /* Call subroutine for moderate red */
      delay(200);                                 /* 0,2 seconds delay */ 
      SUB_NEO_SEND (0, 0, 0);                     /* Call subroutine for off */
      delay(200);                                 /* 0,2 seconds delay */ 
    }
  }
}


/****************************************************************************************/
/* Subroutine send Neopixel values                                                      */
/****************************************************************************************/
void SUB_NEO_SEND (byte color_1, byte color_2, byte color_3)
{
  /* Loop for all pixels */
  for (LOOP_COUNTER = 0; LOOP_COUNTER < NUM_NEO_PIXELS; LOOP_COUNTER++)  
  {                          
    leds[LOOP_COUNTER] = CRGB( color_2, color_1, color_3);          
  }
  FastLED.show(); /* This sends the updated pixel color to the hardware. */ 
}  
